import { Kind, validate, TokenKind, isScalarType, isNonNullType, isListType, isObjectType as isObjectType$1, visit, visitWithTypeInfo, GraphQLObjectType, GraphQLInterfaceType, TypeInfo, isInterfaceType, Source, GraphQLError } from 'graphql';
import { validateSDL } from 'graphql/validation/validate';
import { processImport, parseImportLine } from '@graphql-tools/import';
import { statSync, existsSync, readFileSync } from 'fs';
import { dirname, join, extname, basename, relative, resolve } from 'path';
import { asArray, parseGraphQLSDL } from '@graphql-tools/utils';
import lowerCase from 'lodash.lowercase';
import depthLimit from 'graphql-depth-limit';
import { parseCode } from '@graphql-tools/graphql-tag-pluck';
import { loadConfigSync, GraphQLConfig } from 'graphql-config';
import { CodeFileLoader } from '@graphql-tools/code-file-loader';
import { RuleTester, Linter } from 'eslint';
import { codeFrameColumns } from '@babel/code-frame';

const base = {
    parser: '@graphql-eslint/eslint-plugin',
    plugins: ['@graphql-eslint'],
};

/*
 * 🚨 IMPORTANT! Do not manually modify this file. Run: `yarn generate-configs`
 */
const schemaRecommendedConfig = {
    extends: ['plugin:@graphql-eslint/base'],
    rules: {
        '@graphql-eslint/description-style': 'error',
        '@graphql-eslint/known-argument-names': 'error',
        '@graphql-eslint/known-directives': 'error',
        '@graphql-eslint/known-type-names': 'error',
        '@graphql-eslint/lone-schema-definition': 'error',
        '@graphql-eslint/naming-convention': [
            'error',
            {
                types: 'PascalCase',
                FieldDefinition: 'camelCase',
                InputValueDefinition: 'camelCase',
                Argument: 'camelCase',
                DirectiveDefinition: 'camelCase',
                EnumValueDefinition: 'UPPER_CASE',
                'FieldDefinition[parent.name.value=Query]': {
                    forbiddenPrefixes: ['query', 'get'],
                    forbiddenSuffixes: ['Query'],
                },
                'FieldDefinition[parent.name.value=Mutation]': {
                    forbiddenPrefixes: ['mutation'],
                    forbiddenSuffixes: ['Mutation'],
                },
                'FieldDefinition[parent.name.value=Subscription]': {
                    forbiddenPrefixes: ['subscription'],
                    forbiddenSuffixes: ['Subscription'],
                },
            },
        ],
        '@graphql-eslint/no-case-insensitive-enum-values-duplicates': 'error',
        '@graphql-eslint/no-hashtag-description': 'error',
        '@graphql-eslint/no-typename-prefix': 'error',
        '@graphql-eslint/no-unreachable-types': 'error',
        '@graphql-eslint/provided-required-arguments': 'error',
        '@graphql-eslint/require-deprecation-reason': 'error',
        '@graphql-eslint/require-description': ['error', { types: true, DirectiveDefinition: true }],
        '@graphql-eslint/strict-id-in-types': 'error',
        '@graphql-eslint/unique-directive-names': 'error',
        '@graphql-eslint/unique-directive-names-per-location': 'error',
        '@graphql-eslint/unique-field-definition-names': 'error',
        '@graphql-eslint/unique-operation-types': 'error',
        '@graphql-eslint/unique-type-names': 'error',
    },
};

/*
 * 🚨 IMPORTANT! Do not manually modify this file. Run: `yarn generate-configs`
 */
const schemaAllConfig = {
    extends: ['plugin:@graphql-eslint/base', 'plugin:@graphql-eslint/schema-recommended'],
    rules: {
        '@graphql-eslint/alphabetize': [
            'error',
            {
                fields: ['ObjectTypeDefinition', 'InterfaceTypeDefinition', 'InputObjectTypeDefinition'],
                values: ['EnumTypeDefinition'],
                arguments: ['FieldDefinition', 'Field', 'DirectiveDefinition', 'Directive'],
            },
        ],
        '@graphql-eslint/input-name': 'error',
        '@graphql-eslint/no-scalar-result-type-on-mutation': 'error',
        '@graphql-eslint/require-deprecation-date': 'error',
        '@graphql-eslint/require-field-of-type-query-in-mutation-result': 'error',
    },
};

/*
 * 🚨 IMPORTANT! Do not manually modify this file. Run: `yarn generate-configs`
 */
const operationsRecommendedConfig = {
    extends: ['plugin:@graphql-eslint/base'],
    rules: {
        '@graphql-eslint/executable-definitions': 'error',
        '@graphql-eslint/fields-on-correct-type': 'error',
        '@graphql-eslint/fragments-on-composite-type': 'error',
        '@graphql-eslint/known-argument-names': 'error',
        '@graphql-eslint/known-directives': 'error',
        '@graphql-eslint/known-fragment-names': 'error',
        '@graphql-eslint/known-type-names': 'error',
        '@graphql-eslint/lone-anonymous-operation': 'error',
        '@graphql-eslint/naming-convention': [
            'error',
            {
                VariableDefinition: 'camelCase',
                OperationDefinition: {
                    style: 'PascalCase',
                    forbiddenPrefixes: ['Query', 'Mutation', 'Subscription', 'Get'],
                    forbiddenSuffixes: ['Query', 'Mutation', 'Subscription'],
                },
                FragmentDefinition: { style: 'PascalCase', forbiddenPrefixes: ['Fragment'], forbiddenSuffixes: ['Fragment'] },
            },
        ],
        '@graphql-eslint/no-anonymous-operations': 'error',
        '@graphql-eslint/no-deprecated': 'error',
        '@graphql-eslint/no-duplicate-fields': 'error',
        '@graphql-eslint/no-fragment-cycles': 'error',
        '@graphql-eslint/no-undefined-variables': 'error',
        '@graphql-eslint/no-unused-fragments': 'error',
        '@graphql-eslint/no-unused-variables': 'error',
        '@graphql-eslint/one-field-subscriptions': 'error',
        '@graphql-eslint/overlapping-fields-can-be-merged': 'error',
        '@graphql-eslint/possible-fragment-spread': 'error',
        '@graphql-eslint/provided-required-arguments': 'error',
        '@graphql-eslint/require-id-when-available': 'error',
        '@graphql-eslint/scalar-leafs': 'error',
        '@graphql-eslint/selection-set-depth': ['error', { maxDepth: 7 }],
        '@graphql-eslint/unique-argument-names': 'error',
        '@graphql-eslint/unique-directive-names-per-location': 'error',
        '@graphql-eslint/unique-input-field-names': 'error',
        '@graphql-eslint/unique-variable-names': 'error',
        '@graphql-eslint/value-literals-of-correct-type': 'error',
        '@graphql-eslint/variables-are-input-types': 'error',
        '@graphql-eslint/variables-in-allowed-position': 'error',
    },
};

/*
 * 🚨 IMPORTANT! Do not manually modify this file. Run: `yarn generate-configs`
 */
const operationsAllConfig = {
    extends: ['plugin:@graphql-eslint/base', 'plugin:@graphql-eslint/operations-recommended'],
    rules: {
        '@graphql-eslint/alphabetize': [
            'error',
            {
                selections: ['OperationDefinition', 'FragmentDefinition'],
                variables: ['OperationDefinition'],
                arguments: ['Field', 'Directive'],
            },
        ],
        '@graphql-eslint/match-document-filename': [
            'error',
            { query: 'kebab-case', mutation: 'kebab-case', subscription: 'kebab-case', fragment: 'kebab-case' },
        ],
        '@graphql-eslint/unique-fragment-name': 'error',
        '@graphql-eslint/unique-operation-name': 'error',
    },
};

const configs = {
    base,
    'schema-recommended': schemaRecommendedConfig,
    'schema-all': schemaAllConfig,
    'operations-recommended': operationsRecommendedConfig,
    'operations-all': operationsAllConfig,
};

function requireSiblingsOperations(ruleName, context) {
    if (!context.parserServices) {
        throw new Error(`Rule '${ruleName}' requires 'parserOptions.operations' to be set and loaded. See http://bit.ly/graphql-eslint-operations for more info`);
    }
    if (!context.parserServices.siblingOperations.available) {
        throw new Error(`Rule '${ruleName}' requires 'parserOptions.operations' to be set and loaded. See http://bit.ly/graphql-eslint-operations for more info`);
    }
    return context.parserServices.siblingOperations;
}
function requireGraphQLSchemaFromContext(ruleName, context) {
    if (!context.parserServices) {
        throw new Error(`Rule '${ruleName}' requires 'parserOptions.schema' to be set. See http://bit.ly/graphql-eslint-schema for more info`);
    }
    if (!context.parserServices.hasTypeInfo) {
        throw new Error(`Rule '${ruleName}' requires 'parserOptions.schema' to be set and schema to be loaded. See http://bit.ly/graphql-eslint-schema for more info`);
    }
    return context.parserServices.schema;
}
function requireReachableTypesFromContext(ruleName, context) {
    const schema = requireGraphQLSchemaFromContext(ruleName, context);
    return context.parserServices.reachableTypes(schema);
}
function requireUsedFieldsFromContext(ruleName, context) {
    const schema = requireGraphQLSchemaFromContext(ruleName, context);
    const siblings = requireSiblingsOperations(ruleName, context);
    return context.parserServices.usedFields(schema, siblings);
}
function getLexer(source) {
    // GraphQL v14
    const gqlLanguage = require('graphql/language');
    if (gqlLanguage && gqlLanguage.createLexer) {
        return gqlLanguage.createLexer(source, {});
    }
    // GraphQL v15
    const { Lexer: LexerCls } = require('graphql');
    if (LexerCls && typeof LexerCls === 'function') {
        return new LexerCls(source);
    }
    throw new Error(`Unsupported GraphQL version! Please make sure to use GraphQL v14 or newer!`);
}
function extractTokens(source) {
    const lexer = getLexer(source);
    const tokens = [];
    let token = lexer.advance();
    while (token && token.kind !== '<EOF>') {
        tokens.push({
            type: token.kind,
            loc: {
                start: {
                    line: token.line,
                    column: token.column,
                },
                end: {
                    line: token.line,
                    column: token.column,
                },
            },
            value: token.value,
            range: [token.start, token.end],
        });
        token = lexer.advance();
    }
    return tokens;
}
const normalizePath = (path) => (path || '').replace(/\\/g, '/');
/**
 * https://github.com/prettier/eslint-plugin-prettier/blob/76bd45ece6d56eb52f75db6b4a1efdd2efb56392/eslint-plugin-prettier.js#L71
 * Given a filepath, get the nearest path that is a regular file.
 * The filepath provided by eslint may be a virtual filepath rather than a file
 * on disk. This attempts to transform a virtual path into an on-disk path
 */
const getOnDiskFilepath = (filepath) => {
    try {
        if (statSync(filepath).isFile()) {
            return filepath;
        }
    }
    catch (err) {
        // https://github.com/eslint/eslint/issues/11989
        if (err.code === 'ENOTDIR') {
            return getOnDiskFilepath(dirname(filepath));
        }
    }
    return filepath;
};
const getTypeName = node => ('type' in node ? getTypeName(node.type) : node.name.value);
// Small workaround for the bug in older versions of @graphql-tools/load
// Can be removed after graphql-config bumps to a new version
const loaderCache = new Proxy(Object.create(null), {
    get(cache, key) {
        const value = cache[key];
        if (value) {
            return asArray(value);
        }
        return undefined;
    },
    set(cache, key, value) {
        if (value) {
            cache[key] = asArray(value);
        }
        return true;
    },
});
const TYPES_KINDS = [
    Kind.OBJECT_TYPE_DEFINITION,
    Kind.INTERFACE_TYPE_DEFINITION,
    Kind.ENUM_TYPE_DEFINITION,
    Kind.SCALAR_TYPE_DEFINITION,
    Kind.INPUT_OBJECT_TYPE_DEFINITION,
    Kind.UNION_TYPE_DEFINITION,
];
var CaseStyle;
(function (CaseStyle) {
    CaseStyle["camelCase"] = "camelCase";
    CaseStyle["pascalCase"] = "PascalCase";
    CaseStyle["snakeCase"] = "snake_case";
    CaseStyle["upperCase"] = "UPPER_CASE";
    CaseStyle["kebabCase"] = "kebab-case";
})(CaseStyle || (CaseStyle = {}));
const pascalCase = (str) => lowerCase(str)
    .split(' ')
    .map(word => word.charAt(0).toUpperCase() + word.slice(1))
    .join('');
const camelCase = (str) => {
    const result = pascalCase(str);
    return result.charAt(0).toLowerCase() + result.slice(1);
};
const convertCase = (style, str) => {
    switch (style) {
        case CaseStyle.camelCase:
            return camelCase(str);
        case CaseStyle.pascalCase:
            return pascalCase(str);
        case CaseStyle.snakeCase:
            return lowerCase(str).replace(/ /g, '_');
        case CaseStyle.upperCase:
            return lowerCase(str).replace(/ /g, '_').toUpperCase();
        case CaseStyle.kebabCase:
            return lowerCase(str).replace(/ /g, '-');
    }
};
function getLocation(loc, fieldName = '', offset) {
    const { start } = loc;
    /*
     * ESLint has 0-based column number
     * https://eslint.org/docs/developer-guide/working-with-rules#contextreport
     */
    const { offsetStart = 1, offsetEnd = 1 } = offset !== null && offset !== void 0 ? offset : {};
    return {
        start: {
            line: start.line,
            column: start.column - offsetStart,
        },
        end: {
            line: start.line,
            column: start.column - offsetEnd + fieldName.length,
        },
    };
}

function extractRuleName(stack) {
    const match = (stack || '').match(/validation[/\\]rules[/\\](.*?)\.js:/) || [];
    return match[1] || null;
}
function validateDoc(sourceNode, context, schema, documentNode, rules, ruleName = null) {
    var _a;
    if (((_a = documentNode === null || documentNode === void 0 ? void 0 : documentNode.definitions) === null || _a === void 0 ? void 0 : _a.length) > 0) {
        try {
            const validationErrors = schema ? validate(schema, documentNode, rules) : validateSDL(documentNode, null, rules);
            for (const error of validationErrors) {
                const validateRuleName = ruleName || `[${extractRuleName(error.stack)}]`;
                context.report({
                    loc: getLocation({ start: error.locations[0] }),
                    message: ruleName ? error.message : `${validateRuleName} ${error.message}`,
                });
            }
        }
        catch (e) {
            context.report({
                node: sourceNode,
                message: e.message,
            });
        }
    }
}
const isGraphQLImportFile = rawSDL => {
    const trimmedRawSDL = rawSDL.trimLeft();
    return trimmedRawSDL.startsWith('# import') || trimmedRawSDL.startsWith('#import');
};
const validationToRule = (name, ruleName, docs, getDocumentNode) => {
    var _a;
    let ruleFn = null;
    try {
        ruleFn = require(`graphql/validation/rules/${ruleName}Rule`)[`${ruleName}Rule`];
    }
    catch (e) {
        try {
            ruleFn = require(`graphql/validation/rules/${ruleName}`)[`${ruleName}Rule`];
        }
        catch (e) {
            ruleFn = require('graphql/validation')[`${ruleName}Rule`];
        }
    }
    const requiresSchema = (_a = docs.requiresSchema) !== null && _a !== void 0 ? _a : true;
    return {
        [name]: {
            meta: {
                docs: {
                    recommended: true,
                    ...docs,
                    graphQLJSRuleName: ruleName,
                    requiresSchema,
                    url: `https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/${name}.md`,
                    description: `${docs.description}\n\n> This rule is a wrapper around a \`graphql-js\` validation function. [You can find its source code here](https://github.com/graphql/graphql-js/blob/main/src/validation/rules/${ruleName}Rule.ts).`,
                },
            },
            create(context) {
                return {
                    Document(node) {
                        if (!ruleFn) {
                            // eslint-disable-next-line no-console
                            console.warn(`You rule "${name}" depends on a GraphQL validation rule ("${ruleName}") but it's not available in the "graphql-js" version you are using. Skipping...`);
                            return;
                        }
                        const schema = requiresSchema ? requireGraphQLSchemaFromContext(name, context) : null;
                        let documentNode;
                        const isRealFile = existsSync(context.getFilename());
                        if (isRealFile && getDocumentNode) {
                            documentNode = getDocumentNode(context);
                        }
                        validateDoc(node, context, schema, documentNode || node.rawNode(), [ruleFn], ruleName);
                    },
                };
            },
        },
    };
};
const importFiles = (context) => {
    const code = context.getSourceCode().text;
    if (!isGraphQLImportFile(code)) {
        return null;
    }
    // Import documents because file contains '#import' comments
    return processImport(context.getFilename());
};
const GRAPHQL_JS_VALIDATIONS = Object.assign({}, validationToRule('executable-definitions', 'ExecutableDefinitions', {
    category: 'Operations',
    description: `A GraphQL document is only valid for execution if all definitions are either operation or fragment definitions.`,
}), validationToRule('fields-on-correct-type', 'FieldsOnCorrectType', {
    category: 'Operations',
    description: 'A GraphQL document is only valid if all fields selected are defined by the parent type, or are an allowed meta field such as `__typename`.',
}), validationToRule('fragments-on-composite-type', 'FragmentsOnCompositeTypes', {
    category: 'Operations',
    description: `Fragments use a type condition to determine if they apply, since fragments can only be spread into a composite type (object, interface, or union), the type condition must also be a composite type.`,
}), validationToRule('known-argument-names', 'KnownArgumentNames', {
    category: ['Schema', 'Operations'],
    description: `A GraphQL field is only valid if all supplied arguments are defined by that field.`,
}), validationToRule('known-directives', 'KnownDirectives', {
    category: ['Schema', 'Operations'],
    description: `A GraphQL document is only valid if all \`@directives\` are known by the schema and legally positioned.`,
}), validationToRule('known-fragment-names', 'KnownFragmentNames', {
    category: 'Operations',
    description: `A GraphQL document is only valid if all \`...Fragment\` fragment spreads refer to fragments defined in the same document.`,
    examples: [
        {
            title: 'Incorrect (fragment not defined in the document)',
            code: /* GraphQL */ `
            query {
              user {
                id
                ...UserFields
              }
            }
          `,
        },
        {
            title: 'Correct',
            code: /* GraphQL */ `
            fragment UserFields on User {
              firstName
              lastName
            }

            query {
              user {
                id
                ...UserFields
              }
            }
          `,
        },
        {
            title: 'Correct (existing import to UserFields fragment)',
            code: /* GraphQL */ `
            #import '../UserFields.gql'

            query {
              user {
                id
                ...UserFields
              }
            }
          `,
        },
        {
            title: "False positive case\n\nFor extracting documents from code under the hood we use [graphql-tag-pluck](https://graphql-tools.com/docs/graphql-tag-pluck) that [don't support string interpolation](https://stackoverflow.com/questions/62749847/graphql-codegen-dynamic-fields-with-interpolation/62751311#62751311) for this moment.",
            code: `
            const USER_FIELDS = gql\`
              fragment UserFields on User {
                id
              }
            \`
            
            const GET_USER = /* GraphQL */ \`
              # eslint @graphql-eslint/known-fragment-names: 'error'

              query User {
                user {
                  ...UserFields
                }
              }

              # Will give false positive error 'Unknown fragment "UserFields"'
              \${USER_FIELDS}
            \``,
        },
    ],
}, importFiles), validationToRule('known-type-names', 'KnownTypeNames', {
    category: ['Schema', 'Operations'],
    description: `A GraphQL document is only valid if referenced types (specifically variable definitions and fragment conditions) are defined by the type schema.`,
}), validationToRule('lone-anonymous-operation', 'LoneAnonymousOperation', {
    category: 'Operations',
    description: `A GraphQL document is only valid if when it contains an anonymous operation (the query short-hand) that it contains only that one operation definition.`,
}), validationToRule('lone-schema-definition', 'LoneSchemaDefinition', {
    category: 'Schema',
    description: `A GraphQL document is only valid if it contains only one schema definition.`,
    requiresSchema: false,
}), validationToRule('no-fragment-cycles', 'NoFragmentCycles', {
    category: 'Operations',
    description: `A GraphQL fragment is only valid when it does not have cycles in fragments usage.`,
}), validationToRule('no-undefined-variables', 'NoUndefinedVariables', {
    category: 'Operations',
    description: `A GraphQL operation is only valid if all variables encountered, both directly and via fragment spreads, are defined by that operation.`,
}, importFiles), validationToRule('no-unused-fragments', 'NoUnusedFragments', {
    category: 'Operations',
    description: `A GraphQL document is only valid if all fragment definitions are spread within operations, or spread within other fragments spread within operations.`,
    requiresSiblings: true,
}, context => {
    const siblings = requireSiblingsOperations('no-unused-fragments', context);
    const documents = [...siblings.getOperations(), ...siblings.getFragments()]
        .filter(({ document }) => isGraphQLImportFile(document.loc.source.body))
        .map(({ filePath, document }) => ({
        filePath,
        code: document.loc.source.body,
    }));
    const getParentNode = (filePath) => {
        for (const { filePath: docFilePath, code } of documents) {
            const isFileImported = code
                .split('\n')
                .filter(isGraphQLImportFile)
                .map(line => parseImportLine(line.replace('#', '')))
                .some(o => filePath === join(dirname(docFilePath), o.from));
            if (!isFileImported) {
                continue;
            }
            // Import first file that import this file
            const document = processImport(docFilePath);
            // Import most top file that import this file
            return getParentNode(docFilePath) || document;
        }
        return null;
    };
    return getParentNode(context.getFilename());
}), validationToRule('no-unused-variables', 'NoUnusedVariables', {
    category: 'Operations',
    description: `A GraphQL operation is only valid if all variables defined by an operation are used, either directly or within a spread fragment.`,
}, importFiles), validationToRule('overlapping-fields-can-be-merged', 'OverlappingFieldsCanBeMerged', {
    category: 'Operations',
    description: `A selection set is only valid if all fields (including spreading any fragments) either correspond to distinct response names or can be merged without ambiguity.`,
}), validationToRule('possible-fragment-spread', 'PossibleFragmentSpreads', {
    category: 'Operations',
    description: `A fragment spread is only valid if the type condition could ever possibly be true: if there is a non-empty intersection of the possible parent types, and possible types which pass the type condition.`,
}), validationToRule('possible-type-extension', 'PossibleTypeExtensions', {
    category: 'Schema',
    description: `A type extension is only valid if the type is defined and has the same kind.`,
    requiresSchema: false,
    recommended: false, // TODO: enable after https://github.com/dotansimha/graphql-eslint/issues/787 will be fixed
}), validationToRule('provided-required-arguments', 'ProvidedRequiredArguments', {
    category: ['Schema', 'Operations'],
    description: `A field or directive is only valid if all required (non-null without a default value) field arguments have been provided.`,
}), validationToRule('scalar-leafs', 'ScalarLeafs', {
    category: 'Operations',
    description: `A GraphQL document is valid only if all leaf fields (fields without sub selections) are of scalar or enum types.`,
}), validationToRule('one-field-subscriptions', 'SingleFieldSubscriptions', {
    category: 'Operations',
    description: `A GraphQL subscription is valid only if it contains a single root field.`,
}), validationToRule('unique-argument-names', 'UniqueArgumentNames', {
    category: 'Operations',
    description: `A GraphQL field or directive is only valid if all supplied arguments are uniquely named.`,
}), validationToRule('unique-directive-names', 'UniqueDirectiveNames', {
    category: 'Schema',
    description: `A GraphQL document is only valid if all defined directives have unique names.`,
    requiresSchema: false,
}), validationToRule('unique-directive-names-per-location', 'UniqueDirectivesPerLocation', {
    category: ['Schema', 'Operations'],
    description: `A GraphQL document is only valid if all non-repeatable directives at a given location are uniquely named.`,
}), validationToRule('unique-enum-value-names', 'UniqueEnumValueNames', {
    category: 'Schema',
    description: `A GraphQL enum type is only valid if all its values are uniquely named.`,
    requiresSchema: false,
    recommended: false,
}), validationToRule('unique-field-definition-names', 'UniqueFieldDefinitionNames', {
    category: 'Schema',
    description: `A GraphQL complex type is only valid if all its fields are uniquely named.`,
    requiresSchema: false,
}), validationToRule('unique-input-field-names', 'UniqueInputFieldNames', {
    category: 'Operations',
    description: `A GraphQL input object value is only valid if all supplied fields are uniquely named.`,
    requiresSchema: false,
}), validationToRule('unique-operation-types', 'UniqueOperationTypes', {
    category: 'Schema',
    description: `A GraphQL document is only valid if it has only one type per operation.`,
    requiresSchema: false,
}), validationToRule('unique-type-names', 'UniqueTypeNames', {
    category: 'Schema',
    description: `A GraphQL document is only valid if all defined types have unique names.`,
    requiresSchema: false,
}), validationToRule('unique-variable-names', 'UniqueVariableNames', {
    category: 'Operations',
    description: `A GraphQL operation is only valid if all its variables are uniquely named.`,
}), validationToRule('value-literals-of-correct-type', 'ValuesOfCorrectType', {
    category: 'Operations',
    description: `A GraphQL document is only valid if all value literals are of the type expected at their position.`,
}), validationToRule('variables-are-input-types', 'VariablesAreInputTypes', {
    category: 'Operations',
    description: `A GraphQL operation is only valid if all the variables it defines are of input types (scalar, enum, or input object).`,
}), validationToRule('variables-in-allowed-position', 'VariablesInAllowedPosition', {
    category: 'Operations',
    description: `Variables passed to field arguments conform to type.`,
}));

const ALPHABETIZE = 'ALPHABETIZE';
const fieldsEnum = [
    Kind.OBJECT_TYPE_DEFINITION,
    Kind.INTERFACE_TYPE_DEFINITION,
    Kind.INPUT_OBJECT_TYPE_DEFINITION,
];
const valuesEnum = [Kind.ENUM_TYPE_DEFINITION];
const selectionsEnum = [
    Kind.OPERATION_DEFINITION,
    Kind.FRAGMENT_DEFINITION,
];
const variablesEnum = [Kind.OPERATION_DEFINITION];
const argumentsEnum = [
    Kind.FIELD_DEFINITION,
    Kind.FIELD,
    Kind.DIRECTIVE_DEFINITION,
    Kind.DIRECTIVE,
];
const rule = {
    meta: {
        type: 'suggestion',
        docs: {
            category: ['Schema', 'Operations'],
            description: 'Enforce arrange in alphabetical order for type fields, enum values, input object fields, operation selections and more.',
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/alphabetize.md',
            examples: [
                {
                    title: 'Incorrect',
                    usage: [{ fields: [Kind.OBJECT_TYPE_DEFINITION] }],
                    code: /* GraphQL */ `
            type User {
              password: String
              firstName: String! # should be before "password"
              age: Int # should be before "firstName"
              lastName: String!
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [{ fields: [Kind.OBJECT_TYPE_DEFINITION] }],
                    code: /* GraphQL */ `
            type User {
              age: Int
              firstName: String!
              lastName: String!
              password: String
            }
          `,
                },
                {
                    title: 'Incorrect',
                    usage: [{ values: [Kind.ENUM_TYPE_DEFINITION] }],
                    code: /* GraphQL */ `
            enum Role {
              SUPER_ADMIN
              ADMIN # should be before "SUPER_ADMIN"
              USER
              GOD # should be before "USER"
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [{ values: [Kind.ENUM_TYPE_DEFINITION] }],
                    code: /* GraphQL */ `
            enum Role {
              ADMIN
              GOD
              SUPER_ADMIN
              USER
            }
          `,
                },
                {
                    title: 'Incorrect',
                    usage: [{ selections: [Kind.OPERATION_DEFINITION] }],
                    code: /* GraphQL */ `
            query {
              me {
                firstName
                lastName
                email # should be before "lastName"
              }
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [{ selections: [Kind.OPERATION_DEFINITION] }],
                    code: /* GraphQL */ `
            query {
              me {
                email
                firstName
                lastName
              }
            }
          `,
                },
            ],
            configOptions: {
                schema: [
                    {
                        fields: fieldsEnum,
                        values: valuesEnum,
                        arguments: argumentsEnum,
                    },
                ],
                operations: [
                    {
                        selections: selectionsEnum,
                        variables: variablesEnum,
                        arguments: [Kind.FIELD, Kind.DIRECTIVE],
                    },
                ],
            },
        },
        messages: {
            [ALPHABETIZE]: '"{{ currName }}" should be before "{{ prevName }}"',
        },
        schema: {
            type: 'array',
            minItems: 1,
            maxItems: 1,
            items: {
                type: 'object',
                additionalProperties: false,
                minProperties: 1,
                properties: {
                    fields: {
                        type: 'array',
                        uniqueItems: true,
                        minItems: 1,
                        items: {
                            enum: fieldsEnum,
                        },
                        description: 'Fields of `type`, `interface`, and `input`',
                    },
                    values: {
                        type: 'array',
                        uniqueItems: true,
                        minItems: 1,
                        items: {
                            enum: valuesEnum,
                        },
                        description: 'Values of `enum`',
                    },
                    selections: {
                        type: 'array',
                        uniqueItems: true,
                        minItems: 1,
                        items: {
                            enum: selectionsEnum,
                        },
                        description: 'Selections of operations (`query`, `mutation` and `subscription`) and `fragment`',
                    },
                    variables: {
                        type: 'array',
                        uniqueItems: true,
                        minItems: 1,
                        items: {
                            enum: variablesEnum,
                        },
                        description: 'Variables of operations (`query`, `mutation` and `subscription`)',
                    },
                    arguments: {
                        type: 'array',
                        uniqueItems: true,
                        minItems: 1,
                        items: {
                            enum: argumentsEnum,
                        },
                        description: 'Arguments of fields and directives',
                    },
                },
            },
        },
    },
    create(context) {
        var _a, _b, _c, _d, _e;
        function checkNodes(nodes) {
            let prevName = null;
            for (const node of nodes) {
                const currName = node.name.value;
                if (prevName && prevName > currName) {
                    const isVariableNode = node.kind === Kind.VARIABLE;
                    context.report({
                        loc: getLocation(node.loc, node.name.value, { offsetEnd: isVariableNode ? 0 : 1 }),
                        messageId: ALPHABETIZE,
                        data: isVariableNode
                            ? {
                                currName: `$${currName}`,
                                prevName: `$${prevName}`,
                            }
                            : { currName, prevName },
                    });
                }
                prevName = currName;
            }
        }
        const opts = context.options[0];
        const fields = new Set((_a = opts.fields) !== null && _a !== void 0 ? _a : []);
        const listeners = {};
        const kinds = [
            fields.has(Kind.OBJECT_TYPE_DEFINITION) && [Kind.OBJECT_TYPE_DEFINITION, Kind.OBJECT_TYPE_EXTENSION],
            fields.has(Kind.INTERFACE_TYPE_DEFINITION) && [Kind.INTERFACE_TYPE_DEFINITION, Kind.INTERFACE_TYPE_EXTENSION],
            fields.has(Kind.INPUT_OBJECT_TYPE_DEFINITION) && [
                Kind.INPUT_OBJECT_TYPE_DEFINITION,
                Kind.INPUT_OBJECT_TYPE_EXTENSION,
            ],
        ]
            .filter(Boolean)
            .flat();
        const fieldsSelector = kinds.join(',');
        const hasEnumValues = ((_b = opts.values) === null || _b === void 0 ? void 0 : _b[0]) === Kind.ENUM_TYPE_DEFINITION;
        const selectionsSelector = (_c = opts.selections) === null || _c === void 0 ? void 0 : _c.join(',');
        const hasVariables = ((_d = opts.variables) === null || _d === void 0 ? void 0 : _d[0]) === Kind.OPERATION_DEFINITION;
        const argumentsSelector = (_e = opts.arguments) === null || _e === void 0 ? void 0 : _e.join(',');
        if (fieldsSelector) {
            listeners[fieldsSelector] = (node) => {
                checkNodes(node.fields);
            };
        }
        if (hasEnumValues) {
            const enumValuesSelector = [Kind.ENUM_TYPE_DEFINITION, Kind.ENUM_TYPE_EXTENSION].join(',');
            listeners[enumValuesSelector] = (node) => {
                checkNodes(node.values);
            };
        }
        if (selectionsSelector) {
            listeners[`:matches(${selectionsSelector}) SelectionSet`] = (node) => {
                checkNodes(node.selections
                    // inline fragment don't have name, so we skip them
                    .filter(selection => selection.kind !== Kind.INLINE_FRAGMENT)
                    .map(selection => 
                // sort by alias is field is renamed
                'alias' in selection && selection.alias ? { name: selection.alias } : selection));
            };
        }
        if (hasVariables) {
            listeners.OperationDefinition = (node) => {
                checkNodes(node.variableDefinitions.map(varDef => varDef.variable));
            };
        }
        if (argumentsSelector) {
            listeners[argumentsSelector] = (node) => {
                checkNodes(node.arguments);
            };
        }
        return listeners;
    },
};

const rule$1 = {
    meta: {
        type: 'suggestion',
        docs: {
            examples: [
                {
                    title: 'Incorrect',
                    usage: [{ style: 'inline' }],
                    code: /* GraphQL */ `
            """ Description """
            type someTypeName {
              # ...
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [{ style: 'inline' }],
                    code: /* GraphQL */ `
            " Description "
            type someTypeName {
              # ...
            }
          `,
                },
            ],
            description: 'Require all comments to follow the same style (either block or inline).',
            category: 'Schema',
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/description-style.md',
            recommended: true,
        },
        schema: [
            {
                type: 'object',
                additionalProperties: false,
                properties: {
                    style: {
                        enum: ['block', 'inline'],
                        default: 'block',
                    },
                },
            },
        ],
    },
    create(context) {
        const { style = 'block' } = context.options[0] || {};
        const isBlock = style === 'block';
        return {
            [`.description[type=StringValue][block!=${isBlock}]`](node) {
                context.report({
                    loc: getLocation(node.loc),
                    message: `Unexpected ${isBlock ? 'inline' : 'block'} description`,
                });
            },
        };
    },
};

const isObjectType = (node) => [Kind.OBJECT_TYPE_DEFINITION, Kind.OBJECT_TYPE_EXTENSION].includes(node.type);
const isQueryType = (node) => isObjectType(node) && node.name.value === 'Query';
const isMutationType = (node) => isObjectType(node) && node.name.value === 'Mutation';
const rule$2 = {
    meta: {
        type: 'suggestion',
        docs: {
            description: 'Require mutation argument to be always called "input" and input type to be called Mutation name + "Input".\nUsing the same name for all input parameters will make your schemas easier to consume and more predictable. Using the same name as mutation for InputType will make it easier to find mutations that InputType belongs to.',
            category: 'Schema',
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/input-name.md',
            examples: [
                {
                    title: 'Incorrect',
                    usage: [{ checkInputType: true }],
                    code: /* GraphQL */ `
            type Mutation {
              SetMessage(message: InputMessage): String
            }
          `,
                },
                {
                    title: 'Correct (with checkInputType)',
                    usage: [{ checkInputType: true }],
                    code: /* GraphQL */ `
            type Mutation {
              SetMessage(input: SetMessageInput): String
            }
          `,
                },
                {
                    title: 'Correct (without checkInputType)',
                    usage: [{ checkInputType: false }],
                    code: /* GraphQL */ `
            type Mutation {
              SetMessage(input: AnyInputTypeName): String
            }
          `,
                },
            ],
        },
        schema: [
            {
                type: 'object',
                additionalProperties: false,
                properties: {
                    checkInputType: {
                        type: 'boolean',
                        default: false,
                        description: 'Check that the input type name follows the convention <mutationName>Input',
                    },
                    caseSensitiveInputType: {
                        type: 'boolean',
                        default: true,
                        description: 'Allow for case discrepancies in the input type name',
                    },
                    checkQueries: {
                        type: 'boolean',
                        default: false,
                        description: 'Apply the rule to Queries',
                    },
                    checkMutations: {
                        type: 'boolean',
                        default: true,
                        description: 'Apply the rule to Mutations',
                    },
                },
            },
        ],
    },
    create(context) {
        const options = {
            checkInputType: false,
            caseSensitiveInputType: true,
            checkQueries: false,
            checkMutations: true,
            ...context.options[0],
        };
        const shouldCheckType = node => (options.checkMutations && isMutationType(node)) || (options.checkQueries && isQueryType(node));
        const listeners = {
            'FieldDefinition > InputValueDefinition[name.value!=input]'(node) {
                if (shouldCheckType(node.parent.parent)) {
                    const name = node.name.value;
                    context.report({
                        loc: getLocation(node.loc, name),
                        message: `Input "${name}" should be called "input"`,
                    });
                }
            },
        };
        if (options.checkInputType) {
            listeners['FieldDefinition > InputValueDefinition NamedType'] = (node) => {
                const findInputType = item => {
                    let currentNode = item;
                    while (currentNode.type !== Kind.INPUT_VALUE_DEFINITION) {
                        currentNode = currentNode.parent;
                    }
                    return currentNode;
                };
                const inputValueNode = findInputType(node);
                if (shouldCheckType(inputValueNode.parent.parent)) {
                    const mutationName = `${inputValueNode.parent.name.value}Input`;
                    const name = node.name.value;
                    if ((options.caseSensitiveInputType && node.name.value !== mutationName) ||
                        name.toLowerCase() !== mutationName.toLowerCase()) {
                        context.report({
                            loc: getLocation(node.loc, name),
                            message: `InputType "${name}" name should be "${mutationName}"`,
                        });
                    }
                }
            };
        }
        return listeners;
    },
};

const MATCH_EXTENSION = 'MATCH_EXTENSION';
const MATCH_STYLE = 'MATCH_STYLE';
const ACCEPTED_EXTENSIONS = ['.gql', '.graphql'];
const CASE_STYLES = [
    CaseStyle.camelCase,
    CaseStyle.pascalCase,
    CaseStyle.snakeCase,
    CaseStyle.upperCase,
    CaseStyle.kebabCase,
];
const schemaOption = {
    oneOf: [{ $ref: '#/definitions/asString' }, { $ref: '#/definitions/asObject' }],
};
const rule$3 = {
    meta: {
        type: 'suggestion',
        docs: {
            category: 'Operations',
            description: 'This rule allows you to enforce that the file name should match the operation name.',
            url: `https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/match-document-filename.md`,
            examples: [
                {
                    title: 'Correct',
                    usage: [{ fileExtension: '.gql' }],
                    code: /* GraphQL */ `
            # user.gql
            type User {
              id: ID!
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [{ query: CaseStyle.snakeCase }],
                    code: /* GraphQL */ `
            # user_by_id.gql
            query UserById {
              userById(id: 5) {
                id
                name
                fullName
              }
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [{ fragment: { style: CaseStyle.kebabCase, suffix: '.fragment' } }],
                    code: /* GraphQL */ `
            # user-fields.fragment.gql
            fragment user_fields on User {
              id
              email
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [{ mutation: { style: CaseStyle.pascalCase, suffix: 'Mutation' } }],
                    code: /* GraphQL */ `
            # DeleteUserMutation.gql
            mutation DELETE_USER {
              deleteUser(id: 5)
            }
          `,
                },
                {
                    title: 'Incorrect',
                    usage: [{ fileExtension: '.graphql' }],
                    code: /* GraphQL */ `
            # post.gql
            type Post {
              id: ID!
            }
          `,
                },
                {
                    title: 'Incorrect',
                    usage: [{ query: CaseStyle.pascalCase }],
                    code: /* GraphQL */ `
            # user-by-id.gql
            query UserById {
              userById(id: 5) {
                id
                name
                fullName
              }
            }
          `,
                },
            ],
            configOptions: [
                {
                    query: CaseStyle.kebabCase,
                    mutation: CaseStyle.kebabCase,
                    subscription: CaseStyle.kebabCase,
                    fragment: CaseStyle.kebabCase,
                },
            ],
        },
        messages: {
            [MATCH_EXTENSION]: `File extension "{{ fileExtension }}" don't match extension "{{ expectedFileExtension }}"`,
            [MATCH_STYLE]: `Unexpected filename "{{ filename }}". Rename it to "{{ expectedFilename }}"`,
        },
        schema: {
            definitions: {
                asString: {
                    enum: CASE_STYLES,
                    description: `One of: ${CASE_STYLES.map(t => `\`${t}\``).join(', ')}`,
                },
                asObject: {
                    type: 'object',
                    additionalProperties: false,
                    properties: {
                        style: {
                            enum: CASE_STYLES,
                        },
                        suffix: {
                            type: 'string',
                        },
                    },
                },
            },
            type: 'array',
            maxItems: 1,
            items: {
                type: 'object',
                additionalProperties: false,
                properties: {
                    fileExtension: {
                        enum: ACCEPTED_EXTENSIONS,
                    },
                    query: schemaOption,
                    mutation: schemaOption,
                    subscription: schemaOption,
                    fragment: schemaOption,
                },
            },
        },
    },
    create(context) {
        const options = context.options[0] || {
            fileExtension: null,
        };
        const filePath = context.getFilename();
        const isVirtualFile = !existsSync(filePath);
        if (process.env.NODE_ENV !== 'test' && isVirtualFile) {
            // Skip validation for code files
            return {};
        }
        const fileExtension = extname(filePath);
        const filename = basename(filePath, fileExtension);
        return {
            Document(documentNode) {
                var _a;
                if (options.fileExtension && options.fileExtension !== fileExtension) {
                    context.report({
                        // Report on first character
                        loc: { column: 0, line: 1 },
                        messageId: MATCH_EXTENSION,
                        data: {
                            fileExtension,
                            expectedFileExtension: options.fileExtension,
                        },
                    });
                }
                const firstOperation = documentNode.definitions.find(n => n.kind === Kind.OPERATION_DEFINITION);
                const firstFragment = documentNode.definitions.find(n => n.kind === Kind.FRAGMENT_DEFINITION);
                const node = firstOperation || firstFragment;
                if (!node) {
                    return;
                }
                const docName = (_a = node.name) === null || _a === void 0 ? void 0 : _a.value;
                if (!docName) {
                    return;
                }
                const docType = 'operation' in node ? node.operation : 'fragment';
                let option = options[docType];
                if (!option) {
                    // Config not provided
                    return;
                }
                if (typeof option === 'string') {
                    option = { style: option };
                }
                const expectedExtension = options.fileExtension || fileExtension;
                const expectedFilename = convertCase(option.style, docName) + (option.suffix || '') + expectedExtension;
                const filenameWithExtension = filename + expectedExtension;
                if (expectedFilename !== filenameWithExtension) {
                    context.report({
                        // Report on first character
                        loc: { column: 0, line: 1 },
                        messageId: MATCH_STYLE,
                        data: {
                            expectedFilename,
                            filename: filenameWithExtension,
                        },
                    });
                }
            },
        };
    },
};

const KindToDisplayName = {
    // types
    [Kind.OBJECT_TYPE_DEFINITION]: 'Type',
    [Kind.INTERFACE_TYPE_DEFINITION]: 'Interface',
    [Kind.ENUM_TYPE_DEFINITION]: 'Enumerator',
    [Kind.SCALAR_TYPE_DEFINITION]: 'Scalar',
    [Kind.INPUT_OBJECT_TYPE_DEFINITION]: 'Input type',
    [Kind.UNION_TYPE_DEFINITION]: 'Union',
    // fields
    [Kind.FIELD_DEFINITION]: 'Field',
    [Kind.INPUT_VALUE_DEFINITION]: 'Input property',
    [Kind.ARGUMENT]: 'Argument',
    [Kind.DIRECTIVE_DEFINITION]: 'Directive',
    // rest
    [Kind.ENUM_VALUE_DEFINITION]: 'Enumeration value',
    [Kind.OPERATION_DEFINITION]: 'Operation',
    [Kind.FRAGMENT_DEFINITION]: 'Fragment',
    [Kind.VARIABLE_DEFINITION]: 'Variable',
};
const StyleToRegex = {
    camelCase: /^[a-z][\dA-Za-z]*$/,
    PascalCase: /^[A-Z][\dA-Za-z]*$/,
    snake_case: /^[a-z][\d_a-z]*[\da-z]$/,
    UPPER_CASE: /^[A-Z][\dA-Z_]*[\dA-Z]$/,
};
const ALLOWED_KINDS = Object.keys(KindToDisplayName).sort();
const ALLOWED_STYLES = Object.keys(StyleToRegex);
const schemaOption$1 = {
    oneOf: [{ $ref: '#/definitions/asString' }, { $ref: '#/definitions/asObject' }],
};
const rule$4 = {
    meta: {
        type: 'suggestion',
        docs: {
            description: 'Require names to follow specified conventions.',
            category: ['Schema', 'Operations'],
            recommended: true,
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/naming-convention.md',
            examples: [
                {
                    title: 'Incorrect',
                    usage: [{ types: 'PascalCase', FieldDefinition: 'camelCase' }],
                    code: /* GraphQL */ `
            type user {
              first_name: String!
            }
          `,
                },
                {
                    title: 'Incorrect',
                    usage: [{ FragmentDefinition: { style: 'PascalCase', forbiddenSuffixes: ['Fragment'] } }],
                    code: /* GraphQL */ `
            fragment UserFragment on User {
              # ...
            }
          `,
                },
                {
                    title: 'Incorrect',
                    usage: [{ 'FieldDefinition[parent.name.value=Query]': { forbiddenPrefixes: ['get'] } }],
                    code: /* GraphQL */ `
            type Query {
              getUsers: [User!]!
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [{ types: 'PascalCase', FieldDefinition: 'camelCase' }],
                    code: /* GraphQL */ `
            type User {
              firstName: String
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [{ FragmentDefinition: { style: 'PascalCase', forbiddenSuffixes: ['Fragment'] } }],
                    code: /* GraphQL */ `
            fragment UserFields on User {
              # ...
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [{ 'FieldDefinition[parent.name.value=Query]': { forbiddenPrefixes: ['get'] } }],
                    code: /* GraphQL */ `
            type Query {
              users: [User!]!
            }
          `,
                },
            ],
            configOptions: {
                schema: [
                    {
                        types: 'PascalCase',
                        FieldDefinition: 'camelCase',
                        InputValueDefinition: 'camelCase',
                        Argument: 'camelCase',
                        DirectiveDefinition: 'camelCase',
                        EnumValueDefinition: 'UPPER_CASE',
                        'FieldDefinition[parent.name.value=Query]': {
                            forbiddenPrefixes: ['query', 'get'],
                            forbiddenSuffixes: ['Query'],
                        },
                        'FieldDefinition[parent.name.value=Mutation]': {
                            forbiddenPrefixes: ['mutation'],
                            forbiddenSuffixes: ['Mutation'],
                        },
                        'FieldDefinition[parent.name.value=Subscription]': {
                            forbiddenPrefixes: ['subscription'],
                            forbiddenSuffixes: ['Subscription'],
                        },
                    },
                ],
                operations: [
                    {
                        VariableDefinition: 'camelCase',
                        OperationDefinition: {
                            style: 'PascalCase',
                            forbiddenPrefixes: ['Query', 'Mutation', 'Subscription', 'Get'],
                            forbiddenSuffixes: ['Query', 'Mutation', 'Subscription'],
                        },
                        FragmentDefinition: {
                            style: 'PascalCase',
                            forbiddenPrefixes: ['Fragment'],
                            forbiddenSuffixes: ['Fragment'],
                        },
                    },
                ],
            },
        },
        schema: {
            definitions: {
                asString: {
                    enum: ALLOWED_STYLES,
                    description: `One of: ${ALLOWED_STYLES.map(t => `\`${t}\``).join(', ')}`,
                },
                asObject: {
                    type: 'object',
                    additionalProperties: false,
                    properties: {
                        style: { enum: ALLOWED_STYLES },
                        prefix: { type: 'string' },
                        suffix: { type: 'string' },
                        forbiddenPrefixes: {
                            type: 'array',
                            uniqueItems: true,
                            minItems: 1,
                            items: { type: 'string' },
                        },
                        forbiddenSuffixes: {
                            type: 'array',
                            uniqueItems: true,
                            minItems: 1,
                            items: { type: 'string' },
                        },
                    },
                },
            },
            type: 'array',
            maxItems: 1,
            items: {
                type: 'object',
                additionalProperties: false,
                properties: {
                    types: {
                        ...schemaOption$1,
                        description: `Includes:\n\n${TYPES_KINDS.map(kind => `- \`${kind}\``).join('\n')}`,
                    },
                    ...Object.fromEntries(ALLOWED_KINDS.map(kind => [
                        kind,
                        {
                            ...schemaOption$1,
                            description: `Read more about this kind on [spec.graphql.org](https://spec.graphql.org/October2021/#${kind}).`,
                        },
                    ])),
                    allowLeadingUnderscore: {
                        type: 'boolean',
                        default: false,
                    },
                    allowTrailingUnderscore: {
                        type: 'boolean',
                        default: false,
                    },
                },
                patternProperties: {
                    [`^(${ALLOWED_KINDS.join('|')})(.+)?$`]: schemaOption$1,
                },
                description: [
                    "> It's possible to use a [`selector`](https://eslint.org/docs/developer-guide/selectors) that starts with allowed `ASTNode` names which are described below.",
                    '>',
                    '> Paste or drop code into the editor in [ASTExplorer](https://astexplorer.net) and inspect the generated AST to compose your selector.',
                    '>',
                    '> Example: pattern property `FieldDefinition[parent.name.value=Query]` will match only fields for type `Query`.',
                ].join('\n'),
            },
        },
    },
    create(context) {
        const options = context.options[0] || {};
        const { allowLeadingUnderscore, allowTrailingUnderscore, types, ...restOptions } = options;
        function normalisePropertyOption(kind) {
            const style = restOptions[kind] || types;
            return typeof style === 'object' ? style : { style };
        }
        const checkNode = (selector) => (node) => {
            const { name } = node.kind === Kind.VARIABLE_DEFINITION ? node.variable : node;
            if (!name) {
                return;
            }
            const { prefix, suffix, forbiddenPrefixes, forbiddenSuffixes, style } = normalisePropertyOption(selector);
            const nodeType = KindToDisplayName[node.kind] || node.kind;
            const nodeName = name.value;
            const errorMessage = getErrorMessage();
            if (errorMessage) {
                context.report({
                    loc: getLocation(name.loc, name.value),
                    message: `${nodeType} "${nodeName}" should ${errorMessage}`,
                });
            }
            function getErrorMessage() {
                let name = nodeName;
                if (allowLeadingUnderscore) {
                    name = name.replace(/^_*/, '');
                }
                if (allowTrailingUnderscore) {
                    name = name.replace(/_*$/, '');
                }
                if (prefix && !name.startsWith(prefix)) {
                    return `have "${prefix}" prefix`;
                }
                if (suffix && !name.endsWith(suffix)) {
                    return `have "${suffix}" suffix`;
                }
                const forbiddenPrefix = forbiddenPrefixes === null || forbiddenPrefixes === void 0 ? void 0 : forbiddenPrefixes.find(prefix => name.startsWith(prefix));
                if (forbiddenPrefix) {
                    return `not have "${forbiddenPrefix}" prefix`;
                }
                const forbiddenSuffix = forbiddenSuffixes === null || forbiddenSuffixes === void 0 ? void 0 : forbiddenSuffixes.find(suffix => name.endsWith(suffix));
                if (forbiddenSuffix) {
                    return `not have "${forbiddenSuffix}" suffix`;
                }
                if (style && !ALLOWED_STYLES.includes(style)) {
                    return `be in one of the following options: ${ALLOWED_STYLES.join(', ')}`;
                }
                const caseRegex = StyleToRegex[style];
                if (caseRegex && !caseRegex.test(name)) {
                    return `be in ${style} format`;
                }
            }
        };
        const checkUnderscore = (node) => {
            const name = node.value;
            context.report({
                loc: getLocation(node.loc, name),
                message: `${name.startsWith('_') ? 'Leading' : 'Trailing'} underscores are not allowed`,
            });
        };
        const listeners = {};
        if (!allowLeadingUnderscore) {
            listeners['Name[value=/^_/]:matches([parent.kind!=Field], [parent.kind=Field][parent.alias])'] = checkUnderscore;
        }
        if (!allowTrailingUnderscore) {
            listeners['Name[value=/_$/]:matches([parent.kind!=Field], [parent.kind=Field][parent.alias])'] = checkUnderscore;
        }
        const selectors = new Set([types && TYPES_KINDS, Object.keys(restOptions)].flat().filter(Boolean));
        for (const selector of selectors) {
            listeners[selector] = checkNode(selector);
        }
        return listeners;
    },
};

const NO_ANONYMOUS_OPERATIONS = 'NO_ANONYMOUS_OPERATIONS';
const rule$5 = {
    meta: {
        type: 'suggestion',
        docs: {
            category: 'Operations',
            description: 'Require name for your GraphQL operations. This is useful since most GraphQL client libraries are using the operation name for caching purposes.',
            recommended: true,
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/no-anonymous-operations.md',
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            query {
              # ...
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            query user {
              # ...
            }
          `,
                },
            ],
        },
        messages: {
            [NO_ANONYMOUS_OPERATIONS]: `Anonymous GraphQL operations are forbidden. Please make sure to name your {{ operation }}!`,
        },
        schema: [],
    },
    create(context) {
        return {
            'OperationDefinition[name=undefined]'(node) {
                context.report({
                    loc: getLocation(node.loc, node.operation),
                    data: {
                        operation: node.operation,
                    },
                    messageId: NO_ANONYMOUS_OPERATIONS,
                });
            },
        };
    },
};

const rule$6 = {
    meta: {
        type: 'suggestion',
        docs: {
            url: `https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/no-case-insensitive-enum-values-duplicates.md`,
            category: 'Schema',
            recommended: true,
            description: 'Disallow case-insensitive enum values duplicates.',
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            enum MyEnum {
              Value
              VALUE
              ValuE
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            enum MyEnum {
              Value1
              Value2
              Value3
            }
          `,
                },
            ],
        },
        schema: [],
    },
    create(context) {
        const selector = [Kind.ENUM_TYPE_DEFINITION, Kind.ENUM_TYPE_EXTENSION].join(',');
        return {
            [selector](node) {
                const duplicates = node.values.filter((item, index, array) => array.findIndex(v => v.name.value.toLowerCase() === item.name.value.toLowerCase()) !== index);
                for (const duplicate of duplicates) {
                    const enumName = duplicate.name.value;
                    context.report({
                        loc: getLocation(duplicate.loc, enumName),
                        message: `Case-insensitive enum values duplicates are not allowed! Found: "${enumName}"`,
                    });
                }
            },
        };
    },
};

const NO_DEPRECATED = 'NO_DEPRECATED';
const rule$7 = {
    meta: {
        type: 'suggestion',
        docs: {
            category: 'Operations',
            description: `Enforce that deprecated fields or enum values are not in use by operations.`,
            url: `https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/no-deprecated.md`,
            requiresSchema: true,
            examples: [
                {
                    title: 'Incorrect (field)',
                    code: /* GraphQL */ `
            # In your schema
            type User {
              id: ID!
              name: String! @deprecated(reason: "old field, please use fullName instead")
              fullName: String!
            }

            # Query
            query user {
              user {
                name # This is deprecated, so you'll get an error
              }
            }
          `,
                },
                {
                    title: 'Incorrect (enum value)',
                    code: /* GraphQL */ `
            # In your schema
            type Mutation {
              changeSomething(type: SomeType): Boolean!
            }

            enum SomeType {
              NEW
              OLD @deprecated(reason: "old field, please use NEW instead")
            }

            # Mutation
            mutation {
              changeSomething(
                type: OLD # This is deprecated, so you'll get an error
              ) {
                ...
              }
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            # In your schema
            type User {
              id: ID!
              name: String! @deprecated(reason: "old field, please use fullName instead")
              fullName: String!
            }

            # Query
            query user {
              user {
                id
                fullName
              }
            }
          `,
                },
            ],
            recommended: true,
        },
        messages: {
            [NO_DEPRECATED]: `This {{ type }} is marked as deprecated in your GraphQL schema {{ reason }}`,
        },
        schema: [],
    },
    create(context) {
        return {
            EnumValue(node) {
                requireGraphQLSchemaFromContext('no-deprecated', context);
                const typeInfo = node.typeInfo();
                if (typeInfo && typeInfo.enumValue) {
                    if (typeInfo.enumValue.deprecationReason) {
                        const enumValueName = node.value;
                        context.report({
                            loc: getLocation(node.loc, enumValueName),
                            messageId: NO_DEPRECATED,
                            data: {
                                type: 'enum value',
                                reason: typeInfo.enumValue.deprecationReason ? `(reason: ${typeInfo.enumValue.deprecationReason})` : '',
                            },
                        });
                    }
                }
            },
            Field(node) {
                requireGraphQLSchemaFromContext('no-deprecated', context);
                const typeInfo = node.typeInfo();
                if (typeInfo && typeInfo.fieldDef) {
                    if (typeInfo.fieldDef.deprecationReason) {
                        const fieldName = node.name.value;
                        context.report({
                            loc: getLocation(node.loc, fieldName),
                            messageId: NO_DEPRECATED,
                            data: {
                                type: 'field',
                                reason: typeInfo.fieldDef.deprecationReason ? `(reason: ${typeInfo.fieldDef.deprecationReason})` : '',
                            },
                        });
                    }
                }
            },
        };
    },
};

const NO_DUPLICATE_FIELDS = 'NO_DUPLICATE_FIELDS';
const rule$8 = {
    meta: {
        type: 'suggestion',
        docs: {
            description: `Checks for duplicate fields in selection set, variables in operation definition, or in arguments set of a field.`,
            category: 'Operations',
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/no-duplicate-fields.md',
            recommended: true,
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            query {
              user {
                name
                email
                name # duplicate field
              }
            }
          `,
                },
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            query {
              users(
                first: 100
                skip: 50
                after: "cji629tngfgou0b73kt7vi5jo"
                first: 100 # duplicate argument
              ) {
                id
              }
            }
          `,
                },
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            query (
              $first: Int!
              $first: Int! # duplicate variable
            ) {
              users(first: $first, skip: 50) {
                id
              }
            }
          `,
                },
            ],
        },
        messages: {
            [NO_DUPLICATE_FIELDS]: `{{ type }} "{{ fieldName }}" defined multiple times`,
        },
        schema: [],
    },
    create(context) {
        function checkNode(usedFields, fieldName, type, node) {
            if (usedFields.has(fieldName)) {
                context.report({
                    loc: getLocation((node.kind === Kind.FIELD && node.alias ? node.alias : node).loc, fieldName, {
                        offsetEnd: node.kind === Kind.VARIABLE_DEFINITION ? 0 : 1,
                    }),
                    messageId: NO_DUPLICATE_FIELDS,
                    data: {
                        type,
                        fieldName,
                    },
                });
            }
            else {
                usedFields.add(fieldName);
            }
        }
        return {
            OperationDefinition(node) {
                const set = new Set();
                for (const varDef of node.variableDefinitions) {
                    checkNode(set, varDef.variable.name.value, 'Operation variable', varDef);
                }
            },
            Field(node) {
                const set = new Set();
                for (const arg of node.arguments) {
                    checkNode(set, arg.name.value, 'Field argument', arg);
                }
            },
            SelectionSet(node) {
                var _a;
                const set = new Set();
                for (const selection of node.selections) {
                    if (selection.kind === Kind.FIELD) {
                        checkNode(set, ((_a = selection.alias) === null || _a === void 0 ? void 0 : _a.value) || selection.name.value, 'Field', selection);
                    }
                }
            },
        };
    },
};

const HASHTAG_COMMENT = 'HASHTAG_COMMENT';
const rule$9 = {
    meta: {
        messages: {
            [HASHTAG_COMMENT]: 'Using hashtag (#) for adding GraphQL descriptions is not allowed. Prefer using """ for multiline, or " for a single line description.',
        },
        docs: {
            description: 'Requires to use `"""` or `"` for adding a GraphQL description instead of `#`.\nAllows to use hashtag for comments, as long as it\'s not attached to an AST definition.',
            category: 'Schema',
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/no-hashtag-description.md',
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            # Represents a user
            type User {
              id: ID!
              name: String
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            " Represents a user "
            type User {
              id: ID!
              name: String
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            # This file defines the basic User type.
            # This comment is valid because it's not attached specifically to an AST object.

            " Represents a user "
            type User {
              id: ID! # This one is also valid, since it comes after the AST object
              name: String
            }
          `,
                },
            ],
            recommended: true,
        },
        type: 'suggestion',
        schema: [],
    },
    create(context) {
        return {
            Document(node) {
                const rawNode = node.rawNode();
                let token = rawNode.loc.startToken;
                while (token !== null) {
                    const { kind, prev, next, value, line, column } = token;
                    if (kind === TokenKind.COMMENT && prev && next) {
                        const isEslintComment = value.trimLeft().startsWith('eslint');
                        const linesAfter = next.line - line;
                        if (!isEslintComment && line !== prev.line && next.kind === TokenKind.NAME && linesAfter < 2) {
                            context.report({
                                messageId: HASHTAG_COMMENT,
                                loc: getLocation({ start: { line, column } }),
                            });
                        }
                    }
                    token = next;
                }
            },
        };
    },
};

const ROOT_TYPES = ['mutation', 'subscription'];
const rule$a = {
    meta: {
        type: 'suggestion',
        docs: {
            category: 'Schema',
            description: 'Disallow using root types `mutation` and/or `subscription`.',
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/no-root-type.md',
            requiresSchema: true,
            examples: [
                {
                    title: 'Incorrect',
                    usage: [{ disallow: ['mutation', 'subscription'] }],
                    code: /* GraphQL */ `
            type Mutation {
              createUser(input: CreateUserInput!): User!
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [{ disallow: ['mutation', 'subscription'] }],
                    code: /* GraphQL */ `
            type Query {
              users: [User!]!
            }
          `,
                },
            ],
        },
        schema: {
            type: 'array',
            minItems: 1,
            maxItems: 1,
            items: {
                type: 'object',
                additionalProperties: false,
                required: ['disallow'],
                properties: {
                    disallow: {
                        type: 'array',
                        uniqueItems: true,
                        minItems: 1,
                        items: {
                            enum: ROOT_TYPES,
                        },
                    },
                },
            },
        },
    },
    create(context) {
        const schema = requireGraphQLSchemaFromContext('no-root-type', context);
        const disallow = new Set(context.options[0].disallow);
        const rootTypeNames = [
            disallow.has('mutation') && schema.getMutationType(),
            disallow.has('subscription') && schema.getSubscriptionType(),
        ]
            .filter(Boolean)
            .map(type => type.name);
        if (rootTypeNames.length === 0) {
            return {};
        }
        const selector = [
            `:matches(${Kind.OBJECT_TYPE_DEFINITION}, ${Kind.OBJECT_TYPE_EXTENSION})`,
            '>',
            `${Kind.NAME}[value=/^(${rootTypeNames.join('|')})$/]`,
        ].join(' ');
        return {
            [selector](node) {
                const typeName = node.value;
                context.report({
                    loc: getLocation(node.loc, typeName),
                    message: `Root type "${typeName}" is forbidden`,
                });
            },
        };
    },
};

const rule$b = {
    meta: {
        type: 'suggestion',
        docs: {
            category: 'Schema',
            description: 'Avoid scalar result type on mutation type to make sure to return a valid state.',
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/no-scalar-result-type-on-mutation.md',
            requiresSchema: true,
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            type Mutation {
              createUser: Boolean
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            type Mutation {
              createUser: User!
            }
          `,
                },
            ],
        },
        schema: [],
    },
    create(context) {
        const schema = requireGraphQLSchemaFromContext('no-scalar-result-type-on-mutation', context);
        const mutationType = schema.getMutationType();
        if (!mutationType) {
            return {};
        }
        const selector = [
            `:matches(${Kind.OBJECT_TYPE_DEFINITION}, ${Kind.OBJECT_TYPE_EXTENSION})[name.value=${mutationType.name}]`,
            `> ${Kind.FIELD_DEFINITION} > .gqlType ${Kind.NAME}`,
        ].join(' ');
        return {
            [selector](node) {
                const typeName = node.value;
                const graphQLType = schema.getType(typeName);
                if (isScalarType(graphQLType)) {
                    context.report({
                        loc: getLocation(node.loc, typeName),
                        message: `Unexpected scalar result type "${typeName}"`,
                    });
                }
            },
        };
    },
};

const NO_TYPENAME_PREFIX = 'NO_TYPENAME_PREFIX';
const rule$c = {
    meta: {
        type: 'suggestion',
        docs: {
            category: 'Schema',
            description: 'Enforces users to avoid using the type name in a field name while defining your schema.',
            recommended: true,
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/no-typename-prefix.md',
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            type User {
              userId: ID!
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            type User {
              id: ID!
            }
          `,
                },
            ],
        },
        messages: {
            [NO_TYPENAME_PREFIX]: `Field "{{ fieldName }}" starts with the name of the parent type "{{ typeName }}"`,
        },
        schema: [],
    },
    create(context) {
        return {
            'ObjectTypeDefinition, ObjectTypeExtension, InterfaceTypeDefinition, InterfaceTypeExtension'(node) {
                const typeName = node.name.value;
                const lowerTypeName = typeName.toLowerCase();
                for (const field of node.fields) {
                    const fieldName = field.name.value;
                    if (fieldName.toLowerCase().startsWith(lowerTypeName)) {
                        context.report({
                            data: {
                                fieldName,
                                typeName,
                            },
                            messageId: NO_TYPENAME_PREFIX,
                            loc: getLocation(field.loc, lowerTypeName),
                        });
                    }
                }
            },
        };
    },
};

const UNREACHABLE_TYPE = 'UNREACHABLE_TYPE';
const RULE_NAME = 'no-unreachable-types';
const rule$d = {
    meta: {
        messages: {
            [UNREACHABLE_TYPE]: 'Type "{{ typeName }}" is unreachable',
        },
        docs: {
            description: `Requires all types to be reachable at some level by root level fields.`,
            category: 'Schema',
            url: `https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/${RULE_NAME}.md`,
            requiresSchema: true,
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            type User {
              id: ID!
              name: String
            }

            type Query {
              me: String
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            type User {
              id: ID!
              name: String
            }

            type Query {
              me: User
            }
          `,
                },
            ],
            recommended: true,
        },
        fixable: 'code',
        type: 'suggestion',
        schema: [],
    },
    create(context) {
        const reachableTypes = requireReachableTypesFromContext(RULE_NAME, context);
        function ensureReachability(node) {
            const typeName = node.name.value;
            if (!reachableTypes.has(typeName)) {
                context.report({
                    loc: getLocation(node.name.loc, typeName, { offsetStart: node.kind === Kind.DIRECTIVE_DEFINITION ? 2 : 1 }),
                    messageId: UNREACHABLE_TYPE,
                    data: { typeName },
                    fix: fixer => fixer.remove(node),
                });
            }
        }
        return {
            DirectiveDefinition: ensureReachability,
            ObjectTypeDefinition: ensureReachability,
            ObjectTypeExtension: ensureReachability,
            InterfaceTypeDefinition: ensureReachability,
            InterfaceTypeExtension: ensureReachability,
            ScalarTypeDefinition: ensureReachability,
            ScalarTypeExtension: ensureReachability,
            InputObjectTypeDefinition: ensureReachability,
            InputObjectTypeExtension: ensureReachability,
            UnionTypeDefinition: ensureReachability,
            UnionTypeExtension: ensureReachability,
            EnumTypeDefinition: ensureReachability,
            EnumTypeExtension: ensureReachability,
        };
    },
};

const UNUSED_FIELD = 'UNUSED_FIELD';
const RULE_NAME$1 = 'no-unused-fields';
const rule$e = {
    meta: {
        messages: {
            [UNUSED_FIELD]: `Field "{{fieldName}}" is unused`,
        },
        docs: {
            description: `Requires all fields to be used at some level by siblings operations.`,
            category: 'Schema',
            url: `https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/${RULE_NAME$1}.md`,
            requiresSiblings: true,
            requiresSchema: true,
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            type User {
              id: ID!
              name: String
              someUnusedField: String
            }

            type Query {
              me: User
            }

            query {
              me {
                id
                name
              }
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            type User {
              id: ID!
              name: String
            }

            type Query {
              me: User
            }

            query {
              me {
                id
                name
              }
            }
          `,
                },
            ],
        },
        fixable: 'code',
        type: 'suggestion',
        schema: [],
    },
    create(context) {
        const usedFields = requireUsedFieldsFromContext(RULE_NAME$1, context);
        return {
            FieldDefinition(node) {
                var _a;
                const fieldName = node.name.value;
                const parentTypeName = node.parent.name.value;
                const isUsed = (_a = usedFields[parentTypeName]) === null || _a === void 0 ? void 0 : _a.has(fieldName);
                if (isUsed) {
                    return;
                }
                context.report({
                    loc: getLocation(node.loc, fieldName),
                    messageId: UNUSED_FIELD,
                    data: { fieldName },
                    fix(fixer) {
                        const sourceCode = context.getSourceCode();
                        const tokenBefore = sourceCode.getTokenBefore(node);
                        const tokenAfter = sourceCode.getTokenAfter(node);
                        const isEmptyType = tokenBefore.type === '{' && tokenAfter.type === '}';
                        if (isEmptyType) {
                            // Remove type
                            const { parent } = node;
                            const parentBeforeToken = sourceCode.getTokenBefore(parent);
                            return parentBeforeToken
                                ? fixer.removeRange([parentBeforeToken.range[1], parent.range[1]])
                                : fixer.remove(parent);
                        }
                        // Remove whitespace before token
                        return fixer.removeRange([tokenBefore.range[1], node.range[1]]);
                    },
                });
            },
        };
    },
};

function keyValMap(list, keyFn, valFn) {
    return list.reduce((map, item) => {
        map[keyFn(item)] = valFn(item);
        return map;
    }, Object.create(null));
}
function valueFromNode(valueNode, variables) {
    switch (valueNode.type) {
        case Kind.NULL:
            return null;
        case Kind.INT:
            return parseInt(valueNode.value, 10);
        case Kind.FLOAT:
            return parseFloat(valueNode.value);
        case Kind.STRING:
        case Kind.ENUM:
        case Kind.BOOLEAN:
            return valueNode.value;
        case Kind.LIST:
            return valueNode.values.map(node => valueFromNode(node, variables));
        case Kind.OBJECT:
            return keyValMap(valueNode.fields, field => field.name.value, field => valueFromNode(field.value, variables));
        case Kind.VARIABLE:
            return variables === null || variables === void 0 ? void 0 : variables[valueNode.name.value];
    }
}
function getBaseType(type) {
    if (isNonNullType(type) || isListType(type)) {
        return getBaseType(type.ofType);
    }
    return type;
}
function convertRange(gqlLocation) {
    return [gqlLocation.start, gqlLocation.end];
}
function extractCommentsFromAst(loc) {
    if (!loc) {
        return [];
    }
    const comments = [];
    let token = loc.startToken;
    while (token !== null) {
        const { kind, value, line, column, start, end, next } = token;
        if (kind === TokenKind.COMMENT) {
            comments.push({
                type: 'Block',
                value,
                loc: {
                    start: { line, column },
                    end: { line, column },
                },
                range: [start, end],
            });
        }
        token = next;
    }
    return comments;
}
function convertLocation(gqlLocation) {
    return {
        start: {
            column: gqlLocation.startToken.column,
            line: gqlLocation.startToken.line,
        },
        end: {
            column: gqlLocation.endToken.column,
            line: gqlLocation.endToken.line,
        },
        source: gqlLocation.source.body,
    };
}
function isNodeWithDescription(obj) {
    var _a;
    return (_a = obj) === null || _a === void 0 ? void 0 : _a.description;
}
function convertDescription(node) {
    if (isNodeWithDescription(node)) {
        return [
            {
                type: node.description.block ? 'Block' : 'Line',
                value: node.description.value,
            },
        ];
    }
    return [];
}

// eslint-disable-next-line unicorn/better-regex
const DATE_REGEX = /^\d{2}\/\d{2}\/\d{4}$/;
const MESSAGE_REQUIRE_DATE = 'MESSAGE_REQUIRE_DATE';
const MESSAGE_INVALID_FORMAT = 'MESSAGE_INVALID_FORMAT';
const MESSAGE_INVALID_DATE = 'MESSAGE_INVALID_DATE';
const MESSAGE_CAN_BE_REMOVED = 'MESSAGE_CAN_BE_REMOVED';
const rule$f = {
    meta: {
        type: 'suggestion',
        docs: {
            category: 'Schema',
            description: 'Require deletion date on `@deprecated` directive. Suggest removing deprecated things after deprecated date.',
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/require-deprecation-date.md',
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            type User {
              firstname: String @deprecated
              firstName: String
            }
          `,
                },
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            type User {
              firstname: String @deprecated(reason: "Use 'firstName' instead")
              firstName: String
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            type User {
              firstname: String @deprecated(reason: "Use 'firstName' instead", deletionDate: "25/12/2022")
              firstName: String
            }
          `,
                },
            ],
        },
        messages: {
            [MESSAGE_REQUIRE_DATE]: 'Directive "@deprecated" must have a deletion date',
            [MESSAGE_INVALID_FORMAT]: 'Deletion date must be in format "DD/MM/YYYY"',
            [MESSAGE_INVALID_DATE]: 'Invalid "{{ deletionDate }}" deletion date',
            [MESSAGE_CAN_BE_REMOVED]: '"{{ nodeName }}" сan be removed',
        },
        schema: [
            {
                type: 'object',
                additionalProperties: false,
                properties: {
                    argumentName: {
                        type: 'string',
                    },
                },
            },
        ],
    },
    create(context) {
        return {
            'Directive[name.value=deprecated]'(node) {
                var _a;
                const argName = ((_a = context.options[0]) === null || _a === void 0 ? void 0 : _a.argumentName) || 'deletionDate';
                const deletionDateNode = node.arguments.find(arg => arg.name.value === argName);
                if (!deletionDateNode) {
                    context.report({
                        loc: getLocation(node.loc, node.name.value, { offsetEnd: 0 }),
                        messageId: MESSAGE_REQUIRE_DATE,
                    });
                    return;
                }
                const deletionDate = valueFromNode(deletionDateNode.value);
                const isValidDate = DATE_REGEX.test(deletionDate);
                if (!isValidDate) {
                    context.report({ node: deletionDateNode.value, messageId: MESSAGE_INVALID_FORMAT });
                    return;
                }
                let [day, month, year] = deletionDate.split('/');
                day = day.toString().padStart(2, '0');
                month = month.toString().padStart(2, '0');
                const deletionDateInMS = Date.parse(`${year}-${month}-${day}`);
                if (Number.isNaN(deletionDateInMS)) {
                    context.report({
                        node: deletionDateNode.value,
                        messageId: MESSAGE_INVALID_DATE,
                        data: {
                            deletionDate,
                        },
                    });
                    return;
                }
                const canRemove = Date.now() > deletionDateInMS;
                if (canRemove) {
                    context.report({
                        node,
                        messageId: MESSAGE_CAN_BE_REMOVED,
                        data: {
                            nodeName: node.parent.name.value,
                        },
                    });
                }
            },
        };
    },
};

const rule$g = {
    meta: {
        docs: {
            description: `Require all deprecation directives to specify a reason.`,
            category: 'Schema',
            url: `https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/require-deprecation-reason.md`,
            recommended: true,
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            type MyType {
              name: String @deprecated
            }
          `,
                },
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            type MyType {
              name: String @deprecated(reason: "")
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            type MyType {
              name: String @deprecated(reason: "no longer relevant, please use fullName field")
            }
          `,
                },
            ],
        },
        type: 'suggestion',
        schema: [],
    },
    create(context) {
        return {
            'Directive[name.value=deprecated]'(node) {
                const args = node.arguments || [];
                const reasonArg = args.find(arg => arg.name && arg.name.value === 'reason');
                const value = reasonArg ? String(valueFromNode(reasonArg.value) || '').trim() : null;
                if (!value) {
                    context.report({
                        loc: getLocation(node.loc, node.name.value, { offsetEnd: 0 }),
                        message: 'Directive "@deprecated" must have a reason!',
                    });
                }
            },
        };
    },
};

const REQUIRE_DESCRIPTION_ERROR = 'REQUIRE_DESCRIPTION_ERROR';
const ALLOWED_KINDS$1 = [
    ...TYPES_KINDS,
    Kind.FIELD_DEFINITION,
    Kind.INPUT_VALUE_DEFINITION,
    Kind.ENUM_VALUE_DEFINITION,
    Kind.DIRECTIVE_DEFINITION,
];
const rule$h = {
    meta: {
        docs: {
            category: 'Schema',
            description: 'Enforce descriptions in your type definitions.',
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/require-description.md',
            examples: [
                {
                    title: 'Incorrect',
                    usage: [{ types: true, FieldDefinition: true }],
                    code: /* GraphQL */ `
            type someTypeName {
              name: String
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [{ types: true, FieldDefinition: true }],
                    code: /* GraphQL */ `
            """
            Some type description
            """
            type someTypeName {
              """
              Name description
              """
              name: String
            }
          `,
                },
            ],
            configOptions: [
                {
                    types: true,
                    [Kind.DIRECTIVE_DEFINITION]: true,
                },
            ],
            recommended: true,
        },
        type: 'suggestion',
        messages: {
            [REQUIRE_DESCRIPTION_ERROR]: 'Description is required for nodes of type "{{ nodeType }}"',
        },
        schema: {
            type: 'array',
            minItems: 1,
            maxItems: 1,
            items: {
                type: 'object',
                additionalProperties: false,
                minProperties: 1,
                properties: {
                    types: {
                        type: 'boolean',
                        description: `Includes:\n\n${TYPES_KINDS.map(kind => `- \`${kind}\``).join('\n')}`,
                    },
                    ...Object.fromEntries([...ALLOWED_KINDS$1].sort().map(kind => [
                        kind,
                        {
                            type: 'boolean',
                            description: `Read more about this kind on [spec.graphql.org](https://spec.graphql.org/October2021/#${kind}).`,
                        },
                    ])),
                },
            },
        },
    },
    create(context) {
        const { types, ...restOptions } = context.options[0];
        const kinds = new Set(types ? TYPES_KINDS : []);
        for (const [kind, isEnabled] of Object.entries(restOptions)) {
            if (isEnabled) {
                kinds.add(kind);
            }
            else {
                kinds.delete(kind);
            }
        }
        const selector = [...kinds].join(',');
        return {
            [selector](node) {
                var _a;
                const description = ((_a = node.description) === null || _a === void 0 ? void 0 : _a.value) || '';
                if (description.trim().length === 0) {
                    context.report({
                        loc: getLocation(node.name.loc, node.name.value),
                        messageId: REQUIRE_DESCRIPTION_ERROR,
                        data: {
                            nodeType: node.kind,
                        },
                    });
                }
            },
        };
    },
};

const RULE_NAME$2 = 'require-field-of-type-query-in-mutation-result';
const rule$i = {
    meta: {
        type: 'suggestion',
        docs: {
            category: 'Schema',
            description: 'Allow the client in one round-trip not only to call mutation but also to get a wagon of data to update their application.\n> Currently, no errors are reported for result type `union`, `interface` and `scalar`.',
            url: `https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/${RULE_NAME$2}.md`,
            requiresSchema: true,
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            type User { ... }

            type Mutation {
              createUser: User!
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            type User { ... }

            type Query { ... }

            type CreateUserPayload {
              user: User!
              query: Query!
            }

            type Mutation {
              createUser: CreateUserPayload!
            }
          `,
                },
            ],
        },
        schema: [],
    },
    create(context) {
        const schema = requireGraphQLSchemaFromContext(RULE_NAME$2, context);
        const mutationType = schema.getMutationType();
        const queryType = schema.getQueryType();
        if (!mutationType || !queryType) {
            return {};
        }
        const selector = [
            `:matches(${Kind.OBJECT_TYPE_DEFINITION}, ${Kind.OBJECT_TYPE_EXTENSION})[name.value=${mutationType.name}]`,
            `> ${Kind.FIELD_DEFINITION} > .gqlType ${Kind.NAME}`,
        ].join(' ');
        return {
            [selector](node) {
                const typeName = node.value;
                const graphQLType = schema.getType(typeName);
                if (isObjectType$1(graphQLType)) {
                    const { fields } = graphQLType.astNode;
                    const hasQueryType = fields.some(field => getTypeName(field) === queryType.name);
                    if (!hasQueryType) {
                        context.report({
                            loc: getLocation(node.loc, typeName),
                            message: `Mutation result type "${graphQLType.name}" must contain field of type "${queryType.name}"`,
                        });
                    }
                }
            },
        };
    },
};

function convertToESTree(node, typeInfo) {
    const visitor = { leave: convertNode(typeInfo) };
    return {
        rootTree: visit(node, typeInfo ? visitWithTypeInfo(typeInfo, visitor) : visitor),
        comments: extractCommentsFromAst(node.loc),
    };
}
function hasTypeField(obj) {
    return obj && !!obj.type;
}
/**
 * Strips tokens information from `location` object - this is needed since it's created as linked list in GraphQL-JS,
 * causing eslint to fail on circular JSON
 * @param location
 */
function stripTokens(location) {
    return {
        end: location.end,
        start: location.start,
    };
}
const convertNode = (typeInfo) => (node, key, parent) => {
    const calculatedTypeInfo = typeInfo
        ? {
            argument: typeInfo.getArgument(),
            defaultValue: typeInfo.getDefaultValue(),
            directive: typeInfo.getDirective(),
            enumValue: typeInfo.getEnumValue(),
            fieldDef: typeInfo.getFieldDef(),
            inputType: typeInfo.getInputType(),
            parentInputType: typeInfo.getParentInputType(),
            parentType: typeInfo.getParentType(),
            gqlType: typeInfo.getType(),
        }
        : {};
    const commonFields = {
        typeInfo: () => calculatedTypeInfo,
        leadingComments: convertDescription(node),
        loc: convertLocation(node.loc),
        range: convertRange(node.loc),
    };
    if (hasTypeField(node)) {
        const { type: gqlType, loc: gqlLocation, ...rest } = node;
        const typeFieldSafe = {
            ...rest,
            gqlType,
        };
        const estreeNode = {
            ...typeFieldSafe,
            ...commonFields,
            type: node.kind,
            rawNode: () => {
                if (!parent || key === undefined) {
                    if (node && node.definitions) {
                        return {
                            loc: gqlLocation,
                            kind: Kind.DOCUMENT,
                            definitions: node.definitions.map(d => d.rawNode()),
                        };
                    }
                    return node;
                }
                return parent[key];
            },
            gqlLocation: stripTokens(gqlLocation),
        };
        return estreeNode;
    }
    else {
        const { loc: gqlLocation, ...rest } = node;
        const typeFieldSafe = rest;
        const estreeNode = {
            ...typeFieldSafe,
            ...commonFields,
            type: node.kind,
            rawNode: () => {
                if (!parent || key === undefined) {
                    if (node && node.definitions) {
                        return {
                            loc: gqlLocation,
                            kind: Kind.DOCUMENT,
                            definitions: node.definitions.map(d => d.rawNode()),
                        };
                    }
                    return node;
                }
                return parent[key];
            },
            gqlLocation: stripTokens(gqlLocation),
        };
        return estreeNode;
    }
};

const REQUIRE_ID_WHEN_AVAILABLE = 'REQUIRE_ID_WHEN_AVAILABLE';
const DEFAULT_ID_FIELD_NAME = 'id';
const rule$j = {
    meta: {
        type: 'suggestion',
        docs: {
            category: 'Operations',
            description: 'Enforce selecting specific fields when they are available on the GraphQL type.',
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/require-id-when-available.md',
            requiresSchema: true,
            requiresSiblings: true,
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            # In your schema
            type User {
              id: ID!
              name: String!
            }

            # Query
            query user {
              user {
                name
              }
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            # In your schema
            type User {
              id: ID!
              name: String!
            }

            # Query
            query user {
              user {
                id
                name
              }
            }
          `,
                },
            ],
            recommended: true,
        },
        messages: {
            [REQUIRE_ID_WHEN_AVAILABLE]: `Field "{{ fieldName }}" must be selected when it's available on a type. Please make sure to include it in your selection set!\nIf you are using fragments, make sure that all used fragments {{ checkedFragments }} specifies the field "{{ fieldName }}".`,
        },
        schema: {
            type: 'array',
            maxItems: 1,
            items: {
                type: 'object',
                additionalProperties: false,
                properties: {
                    fieldName: {
                        type: 'string',
                        default: DEFAULT_ID_FIELD_NAME,
                    },
                },
            },
        },
    },
    create(context) {
        return {
            SelectionSet(node) {
                var _a, _b;
                requireGraphQLSchemaFromContext('require-id-when-available', context);
                const siblings = requireSiblingsOperations('require-id-when-available', context);
                const fieldName = (context.options[0] || {}).fieldName || DEFAULT_ID_FIELD_NAME;
                if (!node.selections || node.selections.length === 0) {
                    return;
                }
                const typeInfo = node.typeInfo();
                if (typeInfo && typeInfo.gqlType) {
                    const rawType = getBaseType(typeInfo.gqlType);
                    if (rawType instanceof GraphQLObjectType || rawType instanceof GraphQLInterfaceType) {
                        const fields = rawType.getFields();
                        const hasIdFieldInType = !!fields[fieldName];
                        const checkedFragmentSpreads = new Set();
                        if (hasIdFieldInType) {
                            let found = false;
                            for (const selection of node.selections) {
                                if (selection.kind === 'Field' && selection.name.value === fieldName) {
                                    found = true;
                                }
                                else if (selection.kind === 'InlineFragment') {
                                    found = (((_a = selection.selectionSet) === null || _a === void 0 ? void 0 : _a.selections) || []).some(s => s.kind === 'Field' && s.name.value === fieldName);
                                }
                                else if (selection.kind === 'FragmentSpread') {
                                    const foundSpread = siblings.getFragment(selection.name.value);
                                    if (foundSpread[0]) {
                                        checkedFragmentSpreads.add(foundSpread[0].document.name.value);
                                        found = (((_b = foundSpread[0].document.selectionSet) === null || _b === void 0 ? void 0 : _b.selections) || []).some(s => s.kind === 'Field' && s.name.value === fieldName);
                                    }
                                }
                                if (found) {
                                    break;
                                }
                            }
                            const { parent } = node;
                            const hasIdFieldInInterfaceSelectionSet = parent &&
                                parent.kind === 'InlineFragment' &&
                                parent.parent &&
                                parent.parent.kind === 'SelectionSet' &&
                                parent.parent.selections.some(s => s.kind === 'Field' && s.name.value === fieldName);
                            if (!found && !hasIdFieldInInterfaceSelectionSet) {
                                context.report({
                                    loc: {
                                        start: {
                                            line: node.loc.start.line,
                                            column: node.loc.start.column - 1,
                                        },
                                        end: {
                                            line: node.loc.end.line,
                                            column: node.loc.end.column - 1,
                                        },
                                    },
                                    messageId: REQUIRE_ID_WHEN_AVAILABLE,
                                    data: {
                                        checkedFragments: checkedFragmentSpreads.size === 0 ? '' : `(${Array.from(checkedFragmentSpreads).join(', ')})`,
                                        fieldName,
                                    },
                                });
                            }
                        }
                    }
                }
            },
        };
    },
};

const rule$k = {
    meta: {
        docs: {
            category: 'Operations',
            description: `Limit the complexity of the GraphQL operations solely by their depth. Based on [graphql-depth-limit](https://github.com/stems/graphql-depth-limit).`,
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/selection-set-depth.md',
            requiresSiblings: true,
            examples: [
                {
                    title: 'Incorrect',
                    usage: [{ maxDepth: 1 }],
                    code: `
            query deep2 {
              viewer { # Level 0
                albums { # Level 1
                  title # Level 2
                }
              }
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [{ maxDepth: 4 }],
                    code: `
            query deep2 {
              viewer { # Level 0
                albums { # Level 1
                  title # Level 2
                }
              }
            }
          `,
                },
                {
                    title: 'Correct (ignored field)',
                    usage: [{ maxDepth: 1, ignore: ['albums'] }],
                    code: `
            query deep2 {
              viewer { # Level 0
                albums { # Level 1
                  title # Level 2
                }
              }
            }
          `,
                },
            ],
            recommended: true,
            configOptions: [{ maxDepth: 7 }],
        },
        type: 'suggestion',
        schema: {
            type: 'array',
            minItems: 1,
            maxItems: 1,
            items: {
                type: 'object',
                additionalProperties: false,
                required: ['maxDepth'],
                properties: {
                    maxDepth: {
                        type: 'number',
                    },
                    ignore: {
                        type: 'array',
                        uniqueItems: true,
                        minItems: 1,
                        items: {
                            type: 'string',
                        },
                    },
                },
            },
        },
    },
    create(context) {
        let siblings = null;
        try {
            siblings = requireSiblingsOperations('selection-set-depth', context);
        }
        catch (e) {
            // eslint-disable-next-line no-console
            console.warn(`Rule "selection-set-depth" works best with siblings operations loaded. For more info: http://bit.ly/graphql-eslint-operations`);
        }
        const maxDepth = context.options[0].maxDepth;
        const ignore = context.options[0].ignore || [];
        const checkFn = depthLimit(maxDepth, { ignore });
        return {
            'OperationDefinition, FragmentDefinition'(node) {
                try {
                    const rawNode = node.rawNode();
                    const fragmentsInUse = siblings ? siblings.getFragmentsInUse(rawNode, true) : [];
                    const document = {
                        kind: Kind.DOCUMENT,
                        definitions: [rawNode, ...fragmentsInUse],
                    };
                    checkFn({
                        getDocument: () => document,
                        reportError: (error) => {
                            context.report({
                                loc: getLocation({ start: error.locations[0] }),
                                message: error.message,
                            });
                        },
                    });
                }
                catch (e) {
                    // eslint-disable-next-line no-console
                    console.warn(`Rule "selection-set-depth" check failed due to a missing siblings operations. For more info: http://bit.ly/graphql-eslint-operations`, e);
                }
            },
        };
    },
};

const shouldIgnoreNode = ({ node, exceptions }) => {
    const rawNode = node.rawNode();
    if (exceptions.types && exceptions.types.includes(rawNode.name.value)) {
        return true;
    }
    if (exceptions.suffixes && exceptions.suffixes.some(suffix => rawNode.name.value.endsWith(suffix))) {
        return true;
    }
    return false;
};
const rule$l = {
    meta: {
        type: 'suggestion',
        docs: {
            description: 'Requires output types to have one unique identifier unless they do not have a logical one. Exceptions can be used to ignore output types that do not have unique identifiers.',
            category: 'Schema',
            recommended: true,
            url: 'https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/strict-id-in-types.md',
            examples: [
                {
                    title: 'Incorrect',
                    usage: [{ acceptedIdNames: ['id', '_id'], acceptedIdTypes: ['ID'], exceptions: { suffixes: ['Payload'] } }],
                    code: /* GraphQL */ `
            # Incorrect field name
            type InvalidFieldName {
              key: ID!
            }

            # Incorrect field type
            type InvalidFieldType {
              id: String!
            }

            # Incorrect exception suffix
            type InvalidSuffixResult {
              data: String!
            }

            # Too many unique identifiers. Must only contain one.
            type InvalidFieldName {
              id: ID!
              _id: ID!
            }
          `,
                },
                {
                    title: 'Correct',
                    usage: [
                        {
                            acceptedIdNames: ['id', '_id'],
                            acceptedIdTypes: ['ID'],
                            exceptions: { types: ['Error'], suffixes: ['Payload'] },
                        },
                    ],
                    code: /* GraphQL */ `
            type User {
              id: ID!
            }

            type Post {
              _id: ID!
            }

            type CreateUserPayload {
              data: String!
            }

            type Error {
              message: String!
            }
          `,
                },
            ],
        },
        schema: {
            type: 'array',
            maxItems: 1,
            items: {
                type: 'object',
                additionalProperties: false,
                properties: {
                    acceptedIdNames: {
                        type: 'array',
                        uniqueItems: true,
                        items: {
                            type: 'string',
                        },
                        default: ['id'],
                    },
                    acceptedIdTypes: {
                        type: 'array',
                        uniqueItems: true,
                        items: {
                            type: 'string',
                        },
                        default: ['ID'],
                    },
                    exceptions: {
                        type: 'object',
                        properties: {
                            types: {
                                type: 'array',
                                uniqueItems: true,
                                minItems: 1,
                                description: 'This is used to exclude types with names that match one of the specified values.',
                                items: {
                                    type: 'string',
                                },
                            },
                            suffixes: {
                                type: 'array',
                                uniqueItems: true,
                                minItems: 1,
                                description: 'This is used to exclude types with names with suffixes that match one of the specified values.',
                                items: {
                                    type: 'string',
                                },
                            },
                        },
                    },
                },
            },
        },
    },
    create(context) {
        const options = {
            acceptedIdNames: ['id'],
            acceptedIdTypes: ['ID'],
            exceptions: {},
            ...context.options[0],
        };
        return {
            ObjectTypeDefinition(node) {
                if (shouldIgnoreNode({ node, exceptions: options.exceptions })) {
                    return;
                }
                const validIds = node.fields.filter(field => {
                    const fieldNode = field.rawNode();
                    const isValidIdName = options.acceptedIdNames.includes(fieldNode.name.value);
                    // To be a valid type, it must be non-null and one of the accepted types.
                    let isValidIdType = false;
                    if (fieldNode.type.kind === Kind.NON_NULL_TYPE && fieldNode.type.type.kind === Kind.NAMED_TYPE) {
                        isValidIdType = options.acceptedIdTypes.includes(fieldNode.type.type.name.value);
                    }
                    return isValidIdName && isValidIdType;
                });
                const typeName = node.name.value;
                // Usually, there should be only one unique identifier field per type.
                // Some clients allow multiple fields to be used. If more people need this,
                // we can extend this rule later.
                if (validIds.length !== 1) {
                    context.report({
                        loc: getLocation(node.name.loc, typeName),
                        message: `{{ typeName }} must have exactly one non-nullable unique identifier. Accepted name(s): {{ acceptedNamesString }} ; Accepted type(s): {{ acceptedTypesString }}`,
                        data: {
                            typeName,
                            acceptedNamesString: options.acceptedIdNames.join(','),
                            acceptedTypesString: options.acceptedIdTypes.join(','),
                        },
                    });
                }
            },
        };
    },
};

const RULE_NAME$3 = 'unique-fragment-name';
const UNIQUE_FRAGMENT_NAME = 'UNIQUE_FRAGMENT_NAME';
const checkNode = (context, node, ruleName, messageId) => {
    const documentName = node.name.value;
    const siblings = requireSiblingsOperations(ruleName, context);
    const siblingDocuments = node.kind === Kind.FRAGMENT_DEFINITION ? siblings.getFragment(documentName) : siblings.getOperation(documentName);
    const filepath = context.getFilename();
    const conflictingDocuments = siblingDocuments.filter(f => {
        var _a;
        const isSameName = ((_a = f.document.name) === null || _a === void 0 ? void 0 : _a.value) === documentName;
        const isSamePath = normalizePath(f.filePath) === normalizePath(filepath);
        return isSameName && !isSamePath;
    });
    if (conflictingDocuments.length > 0) {
        context.report({
            messageId,
            data: {
                documentName,
                summary: conflictingDocuments
                    .map(f => `\t${relative(process.cwd(), getOnDiskFilepath(f.filePath))}`)
                    .join('\n'),
            },
            loc: getLocation(node.name.loc, documentName),
        });
    }
};
const rule$m = {
    meta: {
        type: 'suggestion',
        docs: {
            category: 'Operations',
            description: `Enforce unique fragment names across your project.`,
            url: `https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/${RULE_NAME$3}.md`,
            requiresSiblings: true,
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            # user.fragment.graphql
            fragment UserFields on User {
              id
              name
              fullName
            }

            # user-fields.graphql
            fragment UserFields on User {
              id
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            # user.fragment.graphql
            fragment AllUserFields on User {
              id
              name
              fullName
            }

            # user-fields.graphql
            fragment UserFields on User {
              id
            }
          `,
                },
            ],
        },
        messages: {
            [UNIQUE_FRAGMENT_NAME]: 'Fragment named "{{ documentName }}" already defined in:\n{{ summary }}',
        },
        schema: [],
    },
    create(context) {
        return {
            FragmentDefinition(node) {
                checkNode(context, node, RULE_NAME$3, UNIQUE_FRAGMENT_NAME);
            },
        };
    },
};

const RULE_NAME$4 = 'unique-operation-name';
const UNIQUE_OPERATION_NAME = 'UNIQUE_OPERATION_NAME';
const rule$n = {
    meta: {
        type: 'suggestion',
        docs: {
            category: 'Operations',
            description: `Enforce unique operation names across your project.`,
            url: `https://github.com/dotansimha/graphql-eslint/blob/master/docs/rules/${RULE_NAME$4}.md`,
            requiresSiblings: true,
            examples: [
                {
                    title: 'Incorrect',
                    code: /* GraphQL */ `
            # foo.query.graphql
            query user {
              user {
                id
              }
            }

            # bar.query.graphql
            query user {
              me {
                id
              }
            }
          `,
                },
                {
                    title: 'Correct',
                    code: /* GraphQL */ `
            # foo.query.graphql
            query user {
              user {
                id
              }
            }

            # bar.query.graphql
            query me {
              me {
                id
              }
            }
          `,
                },
            ],
        },
        messages: {
            [UNIQUE_OPERATION_NAME]: 'Operation named "{{ documentName }}" already defined in:\n{{ summary }}',
        },
        schema: [],
    },
    create(context) {
        return {
            'OperationDefinition[name!=undefined]'(node) {
                checkNode(context, node, RULE_NAME$4, UNIQUE_OPERATION_NAME);
            },
        };
    },
};

/*
 * 🚨 IMPORTANT! Do not manually modify this file. Run: `yarn generate-configs`
 */
const rules = {
    ...GRAPHQL_JS_VALIDATIONS,
    alphabetize: rule,
    'description-style': rule$1,
    'input-name': rule$2,
    'match-document-filename': rule$3,
    'naming-convention': rule$4,
    'no-anonymous-operations': rule$5,
    'no-case-insensitive-enum-values-duplicates': rule$6,
    'no-deprecated': rule$7,
    'no-duplicate-fields': rule$8,
    'no-hashtag-description': rule$9,
    'no-root-type': rule$a,
    'no-scalar-result-type-on-mutation': rule$b,
    'no-typename-prefix': rule$c,
    'no-unreachable-types': rule$d,
    'no-unused-fields': rule$e,
    'require-deprecation-date': rule$f,
    'require-deprecation-reason': rule$g,
    'require-description': rule$h,
    'require-field-of-type-query-in-mutation-result': rule$i,
    'require-id-when-available': rule$j,
    'selection-set-depth': rule$k,
    'strict-id-in-types': rule$l,
    'unique-fragment-name': rule$m,
    'unique-operation-name': rule$n,
};

const RELEVANT_KEYWORDS = ['gql', 'graphql', '/* GraphQL */'];
function createGraphqlProcessor() {
    const blocksMap = new Map();
    return {
        supportsAutofix: true,
        preprocess(text, filename) {
            const blocks = [];
            blocksMap.set(filename, blocks);
            if (filename && text && RELEVANT_KEYWORDS.some(keyword => text.includes(keyword))) {
                try {
                    const extractedDocuments = parseCode({
                        code: text,
                        filePath: filename,
                        options: {
                            globalGqlIdentifierName: ['gql', 'graphql'],
                            skipIndent: true,
                        },
                    });
                    if ((extractedDocuments === null || extractedDocuments === void 0 ? void 0 : extractedDocuments.length) > 0) {
                        for (const item of extractedDocuments) {
                            blocks.push({
                                filename: `document.graphql`,
                                text: item.content,
                                lineOffset: item.loc.start.line - 1,
                                offset: item.start + 1,
                            });
                        }
                        blocks.push(text);
                        return blocks;
                    }
                }
                catch (e) {
                    // eslint-disable-next-line no-console
                    console.warn(`[graphql-eslint/processor]: Unable to process file "${filename}": `, e);
                }
            }
            return [text];
        },
        postprocess(messageLists, filename) {
            const blocks = blocksMap.get(filename);
            if ((blocks === null || blocks === void 0 ? void 0 : blocks.length) > 0) {
                for (let i = 0; i < messageLists.length; ++i) {
                    const messages = messageLists[i];
                    const block = blocks[i];
                    if (typeof block === 'string') {
                        continue;
                    }
                    const { lineOffset, offset } = block;
                    for (const message of messages) {
                        message.line += lineOffset;
                        if (message.endLine != null) {
                            message.endLine += lineOffset;
                        }
                        if (message.fix && offset !== undefined) {
                            message.fix.range[0] += offset;
                            message.fix.range[1] += offset;
                        }
                    }
                }
            }
            return messageLists.flat();
        },
    };
}

const EXTRACTABLE_FILES_EXTENSIONS = ['.js', '.jsx', '.ts', '.tsx', 'graphql'];
const processor = createGraphqlProcessor();
const processors = EXTRACTABLE_FILES_EXTENSIONS.reduce((prev, ext) => ({ ...prev, [ext]: processor }), {});

const schemaCache = new Map();
function getSchema(options = {}, gqlConfig) {
    const realFilepath = options.filePath ? getOnDiskFilepath(options.filePath) : null;
    const projectForFile = realFilepath ? gqlConfig.getProjectForFile(realFilepath) : gqlConfig.getDefault();
    const schemaKey = asArray(projectForFile.schema)
        .sort()
        .join(',');
    if (!schemaKey) {
        return null;
    }
    let schema = schemaCache.get(schemaKey);
    if (!schema) {
        schema = projectForFile.loadSchemaSync(projectForFile.schema, 'GraphQLSchema', {
            cache: loaderCache,
            ...options.schemaOptions
        });
        schemaCache.set(schemaKey, schema);
    }
    return schema;
}

const handleVirtualPath = (documents) => {
    const filepathMap = Object.create(null);
    return documents.map(source => {
        var _a;
        const { location } = source;
        if (['.gql', '.graphql'].some(extension => location.endsWith(extension))) {
            return source;
        }
        (_a = filepathMap[location]) !== null && _a !== void 0 ? _a : (filepathMap[location] = -1);
        const index = filepathMap[location] += 1;
        return {
            ...source,
            location: resolve(location, `${index}_document.graphql`)
        };
    });
};
const operationsCache = new Map();
const siblingOperationsCache = new Map();
const getSiblings = (filePath, gqlConfig) => {
    const realFilepath = filePath ? getOnDiskFilepath(filePath) : null;
    const projectForFile = realFilepath ? gqlConfig.getProjectForFile(realFilepath) : gqlConfig.getDefault();
    const documentsKey = asArray(projectForFile.documents)
        .sort()
        .join(',');
    if (!documentsKey) {
        return [];
    }
    let siblings = operationsCache.get(documentsKey);
    if (!siblings) {
        const documents = projectForFile.loadDocumentsSync(projectForFile.documents, {
            skipGraphQLImport: true,
            cache: loaderCache
        });
        siblings = handleVirtualPath(documents);
        operationsCache.set(documentsKey, siblings);
    }
    return siblings;
};
function getSiblingOperations(options, gqlConfig) {
    const siblings = getSiblings(options.filePath, gqlConfig);
    if (siblings.length === 0) {
        let printed = false;
        const noopWarn = () => {
            if (!printed) {
                // eslint-disable-next-line no-console
                console.warn(`getSiblingOperations was called without any operations. Make sure to set "parserOptions.operations" to make this feature available!`);
                printed = true;
            }
            return [];
        };
        return {
            available: false,
            getFragments: noopWarn,
            getOperations: noopWarn,
            getFragment: noopWarn,
            getFragmentByType: noopWarn,
            getOperation: noopWarn,
            getOperationByType: noopWarn,
            getFragmentsInUse: noopWarn,
        };
    }
    // Since the siblings array is cached, we can use it as cache key.
    // We should get the same array reference each time we get
    // to this point for the same graphql project
    let siblingOperations = siblingOperationsCache.get(siblings);
    if (!siblingOperations) {
        let fragmentsCache = null;
        const getFragments = () => {
            if (fragmentsCache === null) {
                const result = [];
                for (const source of siblings) {
                    for (const definition of source.document.definitions || []) {
                        if (definition.kind === Kind.FRAGMENT_DEFINITION) {
                            result.push({
                                filePath: source.location,
                                document: definition,
                            });
                        }
                    }
                }
                fragmentsCache = result;
            }
            return fragmentsCache;
        };
        let cachedOperations = null;
        const getOperations = () => {
            if (cachedOperations === null) {
                const result = [];
                for (const source of siblings) {
                    for (const definition of source.document.definitions || []) {
                        if (definition.kind === Kind.OPERATION_DEFINITION) {
                            result.push({
                                filePath: source.location,
                                document: definition,
                            });
                        }
                    }
                }
                cachedOperations = result;
            }
            return cachedOperations;
        };
        const getFragment = (name) => getFragments().filter(f => { var _a; return ((_a = f.document.name) === null || _a === void 0 ? void 0 : _a.value) === name; });
        const collectFragments = (selectable, recursive = true, collected = new Map()) => {
            visit(selectable, {
                FragmentSpread(spread) {
                    const name = spread.name.value;
                    const fragmentInfo = getFragment(name);
                    if (fragmentInfo.length === 0) {
                        // eslint-disable-next-line no-console
                        console.warn(`Unable to locate fragment named "${name}", please make sure it's loaded using "parserOptions.operations"`);
                        return;
                    }
                    const fragment = fragmentInfo[0];
                    const alreadyVisited = collected.has(name);
                    if (!alreadyVisited) {
                        collected.set(name, fragment.document);
                        if (recursive) {
                            collectFragments(fragment.document, recursive, collected);
                        }
                    }
                },
            });
            return collected;
        };
        siblingOperations = {
            available: true,
            getFragments,
            getOperations,
            getFragment,
            getFragmentByType: typeName => getFragments().filter(f => { var _a, _b; return ((_b = (_a = f.document.typeCondition) === null || _a === void 0 ? void 0 : _a.name) === null || _b === void 0 ? void 0 : _b.value) === typeName; }),
            getOperation: name => getOperations().filter(o => { var _a; return ((_a = o.document.name) === null || _a === void 0 ? void 0 : _a.value) === name; }),
            getOperationByType: type => getOperations().filter(o => o.document.operation === type),
            getFragmentsInUse: (selectable, recursive = true) => Array.from(collectFragments(selectable, recursive).values()),
        };
        siblingOperationsCache.set(siblings, siblingOperations);
    }
    return siblingOperations;
}

let graphqlConfig;
function loadGraphqlConfig(options) {
    // We don't want cache config on test environment
    // Otherwise schema and documents will be same for all tests
    if (process.env.NODE_ENV !== 'test' && graphqlConfig) {
        return graphqlConfig;
    }
    const onDiskConfig = options.skipGraphQLConfig
        ? null
        : loadConfigSync({
            throwOnEmpty: false,
            throwOnMissing: false,
            extensions: [addCodeFileLoaderExtension],
        });
    graphqlConfig =
        onDiskConfig ||
            new GraphQLConfig({
                config: options.projects
                    ? { projects: options.projects }
                    : {
                        schema: (options.schema || ''),
                        documents: options.documents || options.operations,
                        extensions: options.extensions,
                        include: options.include,
                        exclude: options.exclude,
                    },
                filepath: 'virtual-config',
            }, [addCodeFileLoaderExtension]);
    return graphqlConfig;
}
const addCodeFileLoaderExtension = api => {
    api.loaders.schema.register(new CodeFileLoader());
    api.loaders.documents.register(new CodeFileLoader());
    return { name: 'graphql-eslint-loaders' };
};

let reachableTypesCache;
function getReachableTypes(schema) {
    // We don't want cache reachableTypes on test environment
    // Otherwise reachableTypes will be same for all tests
    if (process.env.NODE_ENV !== 'test' && reachableTypesCache) {
        return reachableTypesCache;
    }
    const reachableTypes = new Set();
    const collect = (node) => {
        const typeName = getTypeName(node);
        if (reachableTypes.has(typeName)) {
            return;
        }
        reachableTypes.add(typeName);
        const type = schema.getType(typeName) || schema.getDirective(typeName);
        if (isInterfaceType(type)) {
            const { objects, interfaces } = schema.getImplementations(type);
            for (const { astNode } of [...objects, ...interfaces]) {
                visit(astNode, visitor);
            }
        }
        else {
            visit(type.astNode, visitor);
        }
    };
    const visitor = {
        InterfaceTypeDefinition: collect,
        ObjectTypeDefinition: collect,
        InputValueDefinition: collect,
        UnionTypeDefinition: collect,
        FieldDefinition: collect,
        Directive: collect,
        NamedType: collect,
    };
    for (const type of [
        schema,
        schema.getQueryType(),
        schema.getMutationType(),
        schema.getSubscriptionType(),
    ]) {
        if (type) {
            visit(type.astNode, visitor);
        }
    }
    reachableTypesCache = reachableTypes;
    return reachableTypesCache;
}
let usedFieldsCache;
function getUsedFields(schema, operations) {
    // We don't want cache usedFields on test environment
    // Otherwise usedFields will be same for all tests
    if (process.env.NODE_ENV !== 'test' && usedFieldsCache) {
        return usedFieldsCache;
    }
    const usedFields = Object.create(null);
    const typeInfo = new TypeInfo(schema);
    const visitor = visitWithTypeInfo(typeInfo, {
        Field(node) {
            var _a;
            const fieldDef = typeInfo.getFieldDef();
            if (!fieldDef) {
                // skip visiting this node if field is not defined in schema
                return false;
            }
            const parentTypeName = typeInfo.getParentType().name;
            const fieldName = node.name.value;
            (_a = usedFields[parentTypeName]) !== null && _a !== void 0 ? _a : (usedFields[parentTypeName] = new Set());
            usedFields[parentTypeName].add(fieldName);
        },
    });
    const allDocuments = [...operations.getOperations(), ...operations.getFragments()];
    for (const { document } of allDocuments) {
        visit(document, visitor);
    }
    usedFieldsCache = usedFields;
    return usedFieldsCache;
}

function parse(code, options) {
    return parseForESLint(code, options).ast;
}
function parseForESLint(code, options = {}) {
    const gqlConfig = loadGraphqlConfig(options);
    const schema = getSchema(options, gqlConfig);
    const parserServices = {
        hasTypeInfo: schema !== null,
        schema,
        siblingOperations: getSiblingOperations(options, gqlConfig),
        reachableTypes: getReachableTypes,
        usedFields: getUsedFields,
    };
    try {
        const filePath = options.filePath || '';
        const graphqlAst = parseGraphQLSDL(filePath, code, {
            ...options.graphQLParserOptions,
            noLocation: false,
        });
        const { rootTree, comments } = convertToESTree(graphqlAst.document, schema ? new TypeInfo(schema) : null);
        const tokens = extractTokens(new Source(code, filePath));
        return {
            services: parserServices,
            parserServices,
            ast: {
                type: 'Program',
                body: [rootTree],
                sourceType: 'script',
                comments,
                loc: rootTree.loc,
                range: rootTree.range,
                tokens,
            },
        };
    }
    catch (e) {
        // In case of GraphQL parser error, we report it to ESLint as a parser error that matches the requirements
        // of ESLint. This will make sure to display it correctly in IDEs and lint results.
        if (e instanceof GraphQLError) {
            const eslintError = {
                index: e.positions[0],
                lineNumber: e.locations[0].line,
                column: e.locations[0].column,
                message: `[graphql-eslint]: ${e.message}`,
            };
            throw eslintError;
        }
        e.message = `[graphql-eslint]: ${e.message}`;
        throw e;
    }
}

class GraphQLRuleTester extends RuleTester {
    constructor(parserOptions = {}) {
        const config = {
            parser: require.resolve('@graphql-eslint/eslint-plugin'),
            parserOptions: {
                ...parserOptions,
                skipGraphQLConfig: true,
            },
        };
        super(config);
        this.config = config;
    }
    fromMockFile(path) {
        return readFileSync(resolve(__dirname, `../tests/mocks/${path}`), 'utf-8');
    }
    runGraphQLTests(name, rule, tests) {
        const ruleTests = Linter.version.startsWith('8')
            ? tests
            : {
                valid: tests.valid.map(test => {
                    if (typeof test === 'string') {
                        return test;
                    }
                    // eslint-disable-next-line @typescript-eslint/no-unused-vars
                    const { name, ...testCaseOptions } = test;
                    return testCaseOptions;
                }),
                invalid: tests.invalid.map(test => {
                    // ESLint 7 throws an error on CI - Unexpected top-level property "name"
                    // eslint-disable-next-line @typescript-eslint/no-unused-vars
                    const { name, ...testCaseOptions } = test;
                    return testCaseOptions;
                }),
            };
        super.run(name, rule, ruleTests);
        // Skip snapshot testing if `expect` variable is not defined
        if (typeof expect === 'undefined') {
            return;
        }
        const linter = new Linter();
        linter.defineRule(name, rule);
        for (const testCase of tests.invalid) {
            const verifyConfig = getVerifyConfig(name, this.config, testCase);
            defineParser(linter, verifyConfig.parser);
            const { code, filename } = testCase;
            const messages = linter.verify(code, verifyConfig, { filename });
            for (const message of messages) {
                if (message.fatal) {
                    throw new Error(message.message);
                }
                const messageForSnapshot = visualizeEslintMessage(code, message);
                // eslint-disable-next-line no-undef
                expect(messageForSnapshot).toMatchSnapshot();
            }
        }
    }
}
function getVerifyConfig(ruleId, testerConfig, testCase) {
    const { options, parserOptions, parser = testerConfig.parser } = testCase;
    return {
        ...testerConfig,
        parser,
        parserOptions: {
            ...testerConfig.parserOptions,
            ...parserOptions,
        },
        rules: {
            [ruleId]: ['error', ...(Array.isArray(options) ? options : [])],
        },
    };
}
const parsers = new WeakMap();
function defineParser(linter, parser) {
    if (!parser) {
        return;
    }
    if (!parsers.has(linter)) {
        parsers.set(linter, new Set());
    }
    const defined = parsers.get(linter);
    if (!defined.has(parser)) {
        defined.add(parser);
        linter.defineParser(parser, require(parser));
    }
}
function visualizeEslintMessage(text, result) {
    const { line, column, endLine, endColumn, message } = result;
    const location = {
        start: {
            line,
            column,
        },
    };
    if (typeof endLine === 'number' && typeof endColumn === 'number') {
        location.end = {
            line: endLine,
            column: endColumn,
        };
    }
    return codeFrameColumns(text, location, {
        linesAbove: Number.POSITIVE_INFINITY,
        linesBelow: Number.POSITIVE_INFINITY,
        message,
    });
}

export { GraphQLRuleTester, configs, convertDescription, convertLocation, convertRange, convertToESTree, extractCommentsFromAst, getBaseType, isNodeWithDescription, parse, parseForESLint, processors, rules, valueFromNode };
